/* ad_mempool.h

   a data structure for handing out same
   size chunks of memory to adic from a preallocated
   and dynamically expansive pool

   written by mark fashing for boyanna norris and the
   adic project at argonne national laboratory
*/

#ifndef AD_MEMPOOL_H
#define AD_MEMPOOL_H

#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>

#ifndef ad_GRAD_MAX
#define ad_GRAD_MAX 5
#endif

typedef double DERIV_BUCKET[ad_GRAD_MAX];

#define POOL_SIZE_X 500    /* how many elements */
#define POOL_TYPE_X DERIV_BUCKET   /* what data type */
#define POOL_SIZE POOL_SIZE_X * sizeof(POOL_TYPE_X)

struct mempool_struct;
typedef struct mempool_struct mempool2;

struct mempool_struct {
  POOL_TYPE_X pool[POOL_SIZE_X];   /* memory */
  void *freelist;    /* pointers to hand out */
  mempool2 *next;     /* for growing the mempool */
} ;

typedef struct mempool_struct mempool;

/* builds the free list.  assumes that mp has already
   been allocated. */
void mp_init(mempool *mp);

/* hands out a piece of memory from mp */
void *mp_alloc(mempool *mp);

/* hands out a length n array of memory from mp
   by storing an array of pointers to the chunks of
   memory in diarray */
void **mp_allocn(mempool *mp, int n, void **diarray);

/* returns the memory pointed to by ptr to the
   free list of mp */
void mp_free(mempool *mp, void *ptr);

/* returns the memory pointed to by the elements of
   the length n array diarray to the free list of
   mp */
void mp_freen(mempool *mp, int n, void **diarray);

/* dumps the elements of the free list of mp into
   fstream */
void mp_dumpfree(FILE *fstream, mempool *mp);

/* destroys the free list and frees any additional
   memory pools appended to mp.  mp must still be
   deallocated */
void mp_destroy(mempool *mp);

#endif
