#line 533 "modules/ad_grad.w"
    
#line 498 "modules/ad_grad.w"
/*
  THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT
  LAWS OF THE UNITED STATES.  FOR LICENSING INFORMATION CONTACT:

  Christian Bischof or Lucas Roh, Mathematics and Computer Science Division,
  Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL 60439, 
  {bischof,roh}@mcs.anl.gov.
*/



#line 534 "modules/ad_grad.w"
#if !defined(AD_GRAD_H)
#   define AD_GRAD_H
#   if defined(AD_USE_MPI)
#       include "ad_mpi.h"
#   endif
#   include <string.h>
    
#line 510 "modules/ad_grad.w"
#if ad_GRAD_PTR
#include "ad_grad_daxpy.h"
#else
#include "ad_grad_macro_axpys.h"
#endif


#line 541 "modules/ad_grad.w"
    
#line 52 "modules/ad_grad.w"
	

#line 542 "modules/ad_grad.w"
    
#line 58 "modules/ad_grad.w"
    extern int ad_grad_size;
    extern int ad_total_grad_size;   /*only used when generating MPI programs*/
    extern int ad_grad_size_shadow;
    extern int iWiLlNeVeRCoNfLiCt;

#line 543 "modules/ad_grad.w"
#   if defined(__cplusplus)
        extern "C" {
#   endif
    
#line 518 "modules/ad_grad.w"
#if ad_GRAD_PTR

#line 273 "modules/ad_grad.w"
#define AD_VAR_INIT(arg) \
    DERIV_grad(arg) = 0;

#define AD_ARRAY_INIT(arg) \
    memset(arg, '\0', sizeof(arg))

#define AD_VAR_FINAL(arg) \
    ad_grad_daxpy_free(&DERIV_grad(arg))

#define AD_ARRAY_FINAL(arg)\
{   int i;\
    for (i = 0; i < sizeof (arg) / sizeof (arg[0]); i++)\
        ad_grad_daxpy_free(&DERIV_grad(arg[i]));\
}




#line 292 "modules/ad_grad.w"
#define ad_AD_GradInit(n) \
    { \
        ad_grad_size = n; \
        ad_grad_size_shadow = 0; \
        ad_grad_daxpy_init(n); \
    }

#define ad_AD_GradFinal() \
    { \
        ad_grad_daxpy_final(); \
    }

#define ad_AD_ClearGrad(gz)\
        ad_grad_daxpy_free(&gz);

#define ad_AD_CopyGrad(gz,gx) \
        ad_grad_daxpy_copy(gz,gx);

#define ad_AD_GetTotalGradSize() 	ad_grad_size
#define ad_AD_SetTotalGradSize(x)    	/*don't support*/
#define ad_AD_SetIndepDone()
#define ad_AD_ResetIndep() 		ad_AD_ResetShadowVar()

#define ad_AD_ExtractGrad(a, var) \
    { \
        int pOsItIoN; \
	double* gRaDiE = DERIV_grad(var);\
        if (gRaDiE) {\
            for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
                (a)[pOsItIoN] = gRaDiE[pOsItIoN];  \
            }\
        }\
        else {\
            for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
                (a)[pOsItIoN] = 0.0;\
            }\
        }\
    }

#define ad_AD_SetGrad(a, var) \
    { \
        int pOsItIoN; \
        VALIDATE(&DERIV_grad(var));\
        for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
            DERIV_grad(var)[pOsItIoN] = (a)[pOsItIoN];  \
        }\
    }

#define ad_AD_SetGradVal(var, a, i) \
    { \
        int pOsItIoN2; \
        VALIDATE(&DERIV_grad(var));\
        for (pOsItIoN2 = 0; pOsItIoN2 < ad_grad_size; pOsItIoN2++) {\
	    if (pOsItIoN2 == i) {\
                DERIV_grad(var)[pOsItIoN2] = a;  \
            }\
	    else {\
                DERIV_grad(var)[pOsItIoN2] = 0.0;\
            }\
        }\
    }

#define ad_AD_SetIndep(var) \
    { \
        int pOsItIoN = ad_AD_IncrShadowVar(); \
        VALIDATE(&DERIV_grad(var));\
        if (pOsItIoN > ad_GRAD_MAX) {\
            fprintf (stderr, "ad_AD_SetIndep() Error: the number of " \
		     "independent variables %d exceeds the max (GRAD_MAX)!\n",\
		     pOsItIoN); \
            fflush(stderr);  \
            abort();  \
        }\
        ad_AD_SetGradVal(var, 1.0, pOsItIoN); \
    }

#define ad_AD_SetIndepArray(vars, size) \
    { \
        int iWiLlNeVeRCoNfLiCt; \
        for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; \
	       iWiLlNeVeRCoNfLiCt++) { \
            ad_AD_SetIndep((vars)[iWiLlNeVeRCoNfLiCt]); \
        } \
    }






#line 386 "modules/ad_grad.w"
#define AD_VAR_INIT(arg) \
    DERIV_grad(arg) = 0;

#define AD_ARRAY_INIT(arg) \
    memset(arg, '\0', sizeof(arg))

#define AD_VAR_FINAL(arg) \
    ad_grad_daxpy_free(&DERIV_grad(arg))

#define AD_ARRAY_FINAL(arg)\
{   int i;\
    for (i = 0; i < sizeof (arg) / sizeof (arg[0]); i++)\
        ad_grad_daxpy_free(&DERIV_grad(arg[i]));\
}



#line 520 "modules/ad_grad.w"
#elif ad_GRAD_SPARSE_PTR

#line 405 "modules/ad_grad.w"
#define ad_AD_GradInit(n) \
    { \
        ad_grad_size = n; \
        ad_grad_size_shadow = 0; \
        ad_grad_daxpy_init(n); \
    }

#define ad_AD_GradFinal() \
    { \
        ad_grad_daxpy_final(); \
    }

#define ad_AD_ClearGrad(gz)\
        ad_grad_daxpy_free(&gz);

#define ad_AD_CopyGrad(gz,gx) \
        ad_grad_daxpy_copy(gz,gx);

#define ad_AD_GetTotalGradSize() 	ad_grad_size
#define ad_AD_SetTotalGradSize(x)    	/*don't support*/
#define ad_AD_SetIndepDone()
#define ad_AD_ResetIndep() 		ad_AD_ResetShadowVar()

#define ad_AD_ExtractGrad(a, var) \
    { \
        int pOsItIoN; \
	double* gRaDiE = DERIV_grad(var);\
        if (gRaDiE) {\
            for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
                (a)[pOsItIoN] = gRaDiE[pOsItIoN];  \
            }\
        }\
        else {\
            for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
                (a)[pOsItIoN] = 0.0;\
            }\
        }\
    }

#define ad_AD_SetGrad(a, var) \
    { \
        int pOsItIoN; \
        VALIDATE(&DERIV_grad(var));\
        for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
            DERIV_grad(var)[pOsItIoN] = (a)[pOsItIoN];  \
        }\
    }

#define ad_AD_SetGradVal(var, a, i) \
    { \
        int pOsItIoN2; \
        VALIDATE(&DERIV_grad(var));\
        for (pOsItIoN2 = 0; pOsItIoN2 < ad_grad_size; pOsItIoN2++) {\
	    if (pOsItIoN2 == i) {\
                DERIV_grad(var)[pOsItIoN2] = a;  \
            }\
	    else {\
                DERIV_grad(var)[pOsItIoN2] = 0.0;\
            }\
        }\
    }

#define ad_AD_SetIndep(var) \
    { \
        int pOsItIoN = ad_AD_IncrShadowVar(); \
        VALIDATE(&DERIV_grad(var));\
        if (pOsItIoN > ad_GRAD_MAX) {\
            fprintf (stderr, "ad_AD_SetIndep() Error: the number of " \
		     "independent variables %d exceeds the max (GRAD_MAX)!\n",\
		     pOsItIoN); \
            fflush(stderr);  \
            abort();  \
        }\
        ad_AD_SetGradVal(var, 1.0, pOsItIoN); \
    }

#define ad_AD_SetIndepArray(vars, size) \
    { \
        int iWiLlNeVeRCoNfLiCt; \
        for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; \
	       iWiLlNeVeRCoNfLiCt++) { \
            ad_AD_SetIndep((vars)[iWiLlNeVeRCoNfLiCt]); \
        } \
    }





#line 522 "modules/ad_grad.w"
#else
#line 92 "modules/ad_grad.w"
#define ad_AD_GradInitMPI(pargc, pargv) \
    { \
        ad_mpi_init(pargc, pargv, &ad_total_grad_size); \
    }


#line 103 "modules/ad_grad.w"
#define ad_AD_GradFinalMPI() \
    { \
        ad_mpi_final(); \
    }

#line 109 "modules/ad_grad.w"
#define ad_AD_GradFinal() \
    { \
        ad_grad_size = 0; \
    }



#line 150 "modules/ad_grad.w"
#define ad_AD_GradInit(n) \
    { \
	if (n == -1) \
	   ad_grad_size = ad_GRAD_MAX; \
	else \
           ad_grad_size = n; \
        ad_grad_size_shadow = 0; \
    }


#line 161 "modules/ad_grad.w"
#define ad_AD_ClearGrad(gz)\
    {\
        int iWiLlNeVeRCoNfLiCt0;\
        for (iWiLlNeVeRCoNfLiCt0 = 0 ; iWiLlNeVeRCoNfLiCt0 < ad_GRAD_MAX; \
	     iWiLlNeVeRCoNfLiCt0++) {\
	    gz[iWiLlNeVeRCoNfLiCt0] = 0.0;\
        }\
    }
#define ad_AD_CopyGrad(gz,gx) \
    {\
        int iWiLlNeVeRCoNfLiCt0;\
        for (iWiLlNeVeRCoNfLiCt0 = 0 ; iWiLlNeVeRCoNfLiCt0 < ad_GRAD_MAX;\
	     iWiLlNeVeRCoNfLiCt0++) {\
            gz[iWiLlNeVeRCoNfLiCt0] = gx[iWiLlNeVeRCoNfLiCt0];\
        }\
    }


#if defined(AD_USE_MPI)

#   define ad_AD_GetTotalGradSize() ad_total_grad_size
#   define ad_AD_SetTotalGradSize(x) ad_total_grad_size = x
#   define ad_AD_ExtractGrad(a, var) \
    { \
        ad_mpi_gather(DERIV_grad(var), a); \
    }

#   define ad_AD_SetIndepDone() ad_AD_CommitShadowVar()
#   define ad_AD_SetIndep(var) \
    { \
        int pOsItIoN = ad_total_grad_size++; \
        ad_AD_ClearGrad(DERIV_grad(var)); \
        if (pOsItIoN >= ad_mpi_vect_my_disp && \
	    pOsItIoN < ad_mpi_vect_my_disp+ad_mpi_vect_my_len) {\
            int pOsItIoN2 = ad_AD_IncrShadowVar();\
            if (pOsItIoN2 > ad_GRAD_MAX) {\
                fprintf (stderr, "ad_AD_SetIndep() Error: the number of independent variables per processor %d exceeds the max (GRAD_MAX)!\n", pOsItIoN2);\
                fflush(stderr);  \
                abort();  \
            }\
            DERIV_grad(var)[pOsItIoN2] = 1; \
        }\
    }

#   define ad_AD_SetIndepArray(vars, size) \
    { \
        int iWiLlNeVeRCoNfLiCt; \
        for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; \
	       iWiLlNeVeRCoNfLiCt++) { \
            ad_AD_ClearGrad(DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])); \
            int pOsItIoN = ad_total_grad_size++; \
            if (pOsItIoN >= ad_mpi_vect_my_disp && \
	             pOsItIoN < ad_mpi_vect_my_disp+ad_mpi_vect_my_len) {\
                DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])[ad_AD_IncrShadowVar()]=1;\
            } \
        } \
    }


#else /* !defined(AD_USE_MPI)*/

#   define ad_AD_GetTotalGradSize() ad_grad_size
#   define ad_AD_SetTotalGradSize(x) ad_grad_size = x
#   define ad_AD_ExtractGrad(a, var) \
    { \
        int pOsItIoN; \
        for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
            (a)[pOsItIoN] = DERIV_grad(var)[pOsItIoN];  \
        }\
    }
#   define ad_AD_SetGrad(a, var) \
    { \
        int pOsItIoN; \
        for (pOsItIoN = 0; pOsItIoN < ad_grad_size; pOsItIoN++) {\
            DERIV_grad(var)[pOsItIoN] = (a)[pOsItIoN];  \
        }\
    }

#   define ad_AD_SetIndepDone() ad_AD_CommitShadowVar()
#   define ad_AD_ResetIndep() ad_AD_ResetShadowVar()
#   define ad_AD_SetIndep(var) \
    { \
        int pOsItIoN = ad_AD_IncrShadowVar(); \
        if (pOsItIoN > ad_GRAD_MAX) {\
            fprintf (stderr, "ad_AD_SetIndep() Error: the number of " \
		     "independent variables %d exceeds the max (GRAD_MAX)!\n",\
		     pOsItIoN); \
            fflush(stderr);  \
            abort();  \
        }\
        ad_AD_ClearGrad(DERIV_grad(var)); \
        DERIV_grad(var)[pOsItIoN] = 1; \
    }
#   define ad_AD_SetIndepArray(vars, size) \
    { \
        int iWiLlNeVeRCoNfLiCt; \
        for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; \
	       iWiLlNeVeRCoNfLiCt++) { \
            ad_AD_ClearGrad(DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])); \
            DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])[ad_AD_IncrShadowVar()] = 1; \
        } \
    }

#endif /*defined(AD_USE_MPI)*/



#line 524 "modules/ad_grad.w"
#endif


#line 547 "modules/ad_grad.w"
    
#line 118 "modules/ad_grad.w"
    int ad_AD_IncrShadowVar();

#line 127 "modules/ad_grad.w"
    void ad_AD_CommitShadowVar();

#line 136 "modules/ad_grad.w"
    void ad_AD_ResetShadowVar();

#line 548 "modules/ad_grad.w"
#   if defined(__cplusplus)
        }
#  endif
#endif /*AD_GRAD_H*/


