/*
,
  THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT

  LAWS OF THE UNITED STATES.  FOR LICENSING INFORMATION CONTACT:



  Christian Bischof or Lucas Roh, Mathematics and Computer Science Division,

  Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL 60439,

  {bischof,roh}@mcs.anl.gov.

*/

#include <string.h>
#include <stdarg.h>
#include "ad_grad.h" 
#include "ad_grad_daxpy.h"
#include "ad_mempool.h"


void ad_grad_daxpy_init() { 
    ad_adic_deriv_init( ad_grad_size*sizeof(double), 0 );
}

void ad_grad_daxpy_final() { 
    ad_adic_deriv_final();
}


void ad_grad_daxpy_0(DERIV_TYPE* ppz) {
    INVALIDATE(ppz);
}


void ad_grad_daxpy_copy(DERIV_TYPE* ppz, DERIV_TYPE* ppa)
{
  if (IS_ZERO(ppa)) {
    INVALIDATE(ppz);
  } else {
    VALIDATE(ppz);
    memcpy(DERIV_grad(*ppz), DERIV_grad(*ppa), sizeof(double)*ad_grad_size);
  }
}


void ad_grad_daxpy_1(DERIV_TYPE* ppz, double a, DERIV_TYPE* ppa)
{
    if (IS_ZERO(ppa)) {
        INVALIDATE(ppz);
    }
    else {
        DAXPY1(ppz,a,ppa);
    }
}


void ad_grad_daxpy_2(DERIV_TYPE* ppz, double a, DERIV_TYPE* ppa, 
                     double b, DERIV_TYPE* ppb)
{
    if (IS_ZERO(ppa)) {
        if (IS_ZERO(ppb)) {
            INVALIDATE(ppz);
        }
        else {
            DAXPY1(ppz,b,ppb);
        }
    }
    else if (IS_ZERO(ppb)) {
        DAXPY1(ppz,a,ppa);
    }
    else {
        DAXPY2(ppz,a,ppa,b,ppb);
    }
}

void ad_grad_daxpy_3(DERIV_TYPE* ppz, double a, DERIV_TYPE* ppa, 
                     double b, DERIV_TYPE* ppb, double c, DERIV_TYPE* ppc)
{
  if (IS_ZERO(ppa)) {
    if (IS_ZERO(ppb)) {
      if (IS_ZERO(ppc)) { 
	INVALIDATE(ppz);
      } else {      
	DAXPY1(ppz,c,ppc);
      }
    }
    else if (IS_ZERO(ppc)) {
      DAXPY1(ppz,b,ppb);
    } else { 
      DAXPY2(ppz,b,ppb,c,ppc);
    }
  } else if (IS_ZERO(ppb)) {
    if (IS_ZERO(ppc)) {
      DAXPY1(ppz,a,ppa);
    } else { 
      DAXPY2(ppz,a,ppa,c,ppc);
    }
  } else if (IS_ZERO(ppc)) {
    DAXPY2(ppz,a,ppa,b,ppb); 
  } else {
    DAXPY3(ppz,a,ppa,b,ppb,c,ppc);
  }
}

void ad_grad_daxpy_4(DERIV_TYPE* ppz, double a, DERIV_TYPE* ppa, 
                     double b, DERIV_TYPE* ppb, double c, DERIV_TYPE* ppc,
		     double d, DERIV_TYPE* ppd) { 

  double *pz; int i, flag = 0;
  SET_ZERO_FLAG(flag, ppa, 0);
  SET_ZERO_FLAG(flag, ppb, 1);
  SET_ZERO_FLAG(flag, ppc, 2);
  SET_ZERO_FLAG(flag, ppd, 3);
  switch (flag) 
    {
    case 0:  DAXPY4(ppz,a,ppa,b,ppb,c,ppc,d,ppd); break;   
    case 1:  DAXPY3(ppz,b,ppb,c,ppc,d,ppd); break;
    case 2:  DAXPY3(ppz,a,ppa,c,ppc,d,ppd); break;
    case 3:  DAXPY2(ppz,c,ppc,d,ppd); break;
    case 4:  DAXPY3(ppz,a,ppa,b,ppb,d,ppd); break;
    case 5:  DAXPY2(ppz,b,ppb,d,ppd); break;
    case 6:  DAXPY2(ppz,a,ppa,d,ppd); break;
    case 7:  DAXPY1(ppz,d,ppd); break;
    case 8:  DAXPY3(ppz,a,ppa,b,ppb,c,ppc); break;
    case 9:  DAXPY2(ppz,b,ppb,c,ppc); break;
    case 10: DAXPY2(ppz,a,ppa,c,ppc); break;
    case 11: DAXPY1(ppz,c,ppc); break;
    case 12: DAXPY2(ppz,a,ppa,b,ppb); break;
    case 13: DAXPY1(ppz,b,ppb); break;
    case 14: DAXPY1(ppz,a,ppa); break;
    case 15:
      INVALIDATE(ppz);
    }
}

void ad_grad_daxpy_5(DERIV_TYPE* ppz, double a, DERIV_TYPE* ppa, 
                     double b, DERIV_TYPE* ppb, double c, DERIV_TYPE* ppc,
		     double d, DERIV_TYPE* ppd, double e, DERIV_TYPE* ppe) {

  double *pz; int i, flag = 0;
  SET_ZERO_FLAG(flag, ppa, 0);
  SET_ZERO_FLAG(flag, ppb, 1);
  SET_ZERO_FLAG(flag, ppc, 2);
  SET_ZERO_FLAG(flag, ppd, 3);
  SET_ZERO_FLAG(flag, ppe, 4);
  switch (flag) 
    {
    case 0:  DAXPY5(ppz,a,ppa,b,ppb,c,ppc,d,ppd,e,ppe); break;
    case 1:  DAXPY4(ppz,b,ppb,c,ppc,d,ppd,e,ppe); break;
    case 2:  DAXPY4(ppz,a,ppa,c,ppc,d,ppd,e,ppe); break;
    case 3:  DAXPY3(ppz,c,ppc,d,ppd,e,ppe); break;
    case 4:  DAXPY4(ppz,a,ppa,b,ppb,d,ppd,e,ppe); break;
    case 5:  DAXPY3(ppz,b,ppb,d,ppd,e,ppe); break;
    case 6:  DAXPY3(ppz,a,ppa,d,ppd,e,ppe); break;
    case 7:  DAXPY2(ppz,d,ppd,e,ppe); break;
    case 8:  DAXPY4(ppz,a,ppa,b,ppb,c,ppc,e,ppe); break;
    case 9:  DAXPY3(ppz,b,ppb,c,ppc,e,ppe); break;
    case 10: DAXPY3(ppz,a,ppa,c,ppc,e,ppe); break;
    case 11: DAXPY2(ppz,c,ppc,e,ppe); break;
    case 12: DAXPY3(ppz,a,ppa,b,ppb,e,ppe); break;
    case 13: DAXPY2(ppz,b,ppb,e,ppe); break;
    case 14: DAXPY2(ppz,a,ppa,e,ppe); break;
    case 15: DAXPY1(ppz,e,ppe); break;
    case 16: DAXPY4(ppz,a,ppa,b,ppb,c,ppc,d,ppd); break;   
    case 17: DAXPY3(ppz,b,ppb,c,ppc,d,ppd); break;
    case 18: DAXPY3(ppz,a,ppa,c,ppc,d,ppd); break;
    case 19: DAXPY2(ppz,c,ppc,d,ppd); break;
    case 20: DAXPY3(ppz,a,ppa,b,ppb,d,ppd); break;
    case 21: DAXPY2(ppz,b,ppb,d,ppd); break;
    case 22: DAXPY2(ppz,a,ppa,d,ppd); break;
    case 23: DAXPY1(ppz,d,ppd); break;
    case 24: DAXPY3(ppz,a,ppa,b,ppb,c,ppc); break;
    case 25: DAXPY2(ppz,b,ppb,c,ppc); break;
    case 26: DAXPY2(ppz,a,ppa,c,ppc); break;
    case 27: DAXPY1(ppz,c,ppc); break;
    case 28: DAXPY2(ppz,a,ppa,b,ppb); break;
    case 29: DAXPY1(ppz,b,ppb); break;
    case 30: DAXPY1(ppz,a,ppa); break;
    case 31:
      INVALIDATE(ppz);
  }
}

void ad_grad_daxpy_n(int n, DERIV_TYPE *ppz, ...)
{
  static double       alphas[100], *z;
  static DERIV_TYPE*  grads[100];
  int                 i, j, count = 0;
  va_list             parg;
  
  va_start(parg, ppz);
  for (i = 0; i < n; i++) {
    alphas[count] = va_arg(parg, double);
    grads[count]  = (DERIV_TYPE*)va_arg(parg, DERIV_TYPE*);
    if (!IS_ZERO(grads[count])) {
      count++;
    }
  }
  va_end(parg);
  
  switch (count) 
    {
    case 0:
      INVALIDATE(ppz);
      break;
      
    case 1: DAXPY1(ppz,alphas[0],grads[0]); break; 
    case 2: DAXPY2(ppz,alphas[0],grads[0],alphas[1],grads[1]); break;    
    case 3: DAXPY3(ppz,alphas[0],grads[0],alphas[1],grads[1],alphas[2],grads[2]); break;
    case 4: DAXPY4(ppz,alphas[0],grads[0],alphas[1],grads[1],alphas[2],grads[2],
		   alphas[3],grads[3]); break;
    case 5: DAXPY5(ppz,alphas[0],grads[0],alphas[1],grads[1],alphas[2],grads[2],
		   alphas[3],grads[3],alphas[4],grads[4]); break;
    default:
      DAXPY5(ppz,alphas[0],grads[0],alphas[1],grads[1],alphas[2],grads[2],
	     alphas[3],grads[3],alphas[4],grads[4]); 
      z = DERIV_grad(*ppz); 
      for (j = 5; j < count; j++) {
	double    a = alphas[j], *grad = DERIV_grad(*grads[j]);
	for (i = 0; i < ad_grad_size; i++) {
	  z[i] += a*grad[i];
	}
      }
      break;
    
    }
}
