
#line 704 "ad_adj.w"
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include "ad_adj.h"


#line 97 "ad_adj.w"
enum {
   OP_ASSIGN_ZERO = 1,
   OP_ASSIGN_COPY,
   OP_ASSIGN_ONE,
/*
   OP_ASSIGN_TWO,
   OP_ASSIGN_THREE,
 */
   OP_ASSIGN,
};

#define STACK_OFFSET 	1024



#line 195 "ad_adj.w"
#define MODIFIER_OFFSET	3
#define MODIFIER_MASK	((1<<MODIFIER_OFFSET) - 1)
#define OPCODE_BUFSIZE	102400

static FILE* pfd_op = 0;
static unsigned char bufOp[OPCODE_BUFSIZE];
static int countOp = 0;



#line 268 "ad_adj.w"
#define ADDR_BUFSIZE	102400

static FILE* pfd_addr = 0;
static double* bufAddr[ADDR_BUFSIZE];
static int countAddr = 0;



#line 342 "ad_adj.w"
#define ALPHA_BUFSIZE	102400

static FILE* pfd_alpha = 0;
static double bufAlpha[ALPHA_BUFSIZE];
static int countAlpha = 0;



#line 531 "ad_adj.w"
static char* stack_low = 0;
static char* stack_high = 0;



#line 558 "ad_adj.w"
static char* heap_low = 0;
static char* heap_high = 0;


#line 127 "ad_adj.w"
FILE* ad_tape_open(const char* name)
{
    FILE* pfd = fopen(name, "wb+");
    if (!pfd) {
	printf("Cannot open file: %s\n", name);
	exit(1);
    }
    return pfd;
}


void ad_tape_rev_start(FILE* pfd)
{
    if (fseek(pfd, 0L, SEEK_END)) {
	printf("Cannot seek to end\n");
	exit(1);
    }
}


void ad_tape_write(FILE* pfd, void* buf, int bytes)
{
    int n = fwrite(buf, bytes, 1, pfd);
    if (n != 1) {
	printf("Cannot write file\n");
	exit(1);
    }
}


int ad_tape_read(FILE* pfd, void* buf, int bytes)
{
    int n, err;
    if ((n = ftell(pfd)) < bytes) {
        bytes = n;
    }
    if (fseek(pfd, (long)-bytes, SEEK_CUR)) {
	printf("Cannot seek to end");
	exit(1);
    }
    n = fread(buf, 1, bytes, pfd);
    fseek(pfd, (long)-bytes, SEEK_CUR);
    return n;
}



void tape_read_forward();
void tape_read_reverse();
void tape_end();
void tape_begin();




#line 206 "ad_adj.w"
void ad_tape_open_op()
{
    countOp = 0;
    pfd_op = ad_tape_open(".adic.tape.op");
}

void ad_tape_start_op()
{
    countOp = 0;
}

void ad_tape_rev_start_op()
{
    countOp = 0;
    ad_tape_rev_start(pfd_op);
}

void ad_tape_put_op(int op, int count)
{
    bufOp[countOp++] = (op | count<<MODIFIER_OFFSET);
    if (countOp >= OPCODE_BUFSIZE) {
	ad_tape_write(pfd_op, bufOp, OPCODE_BUFSIZE);
	countOp = 0;
    }
}

void ad_tape_stop_op()
{
    if (countOp > 0) {
	ad_tape_write(pfd_op, bufOp, countOp);
	countOp = 0;
    }
    fflush(pfd_op);
}


int ad_tape_get_op(int* count)
{
    int tmp;
    if (countOp <= 0) {
	countOp = ad_tape_read(pfd_op, bufOp, OPCODE_BUFSIZE);
    }
    tmp = bufOp[--countOp];
    *count = tmp>>MODIFIER_OFFSET;
    return tmp & MODIFIER_MASK;
}
    


#line 277 "ad_adj.w"
void ad_tape_open_addr()
{
    countAddr = 0;
    pfd_addr = ad_tape_open(".adic.tape.addr");
}

void ad_tape_start_addr()
{
    countAddr = 0;
}

void ad_tape_rev_start_addr()
{
    countAddr = 0;
    ad_tape_rev_start(pfd_addr);
}

void ad_tape_put_addr(double* pa)
{
    /*    printf("put:%d ", pa);*/
    bufAddr[countAddr++] = pa;
    if (countAddr >= ADDR_BUFSIZE) {
	ad_tape_write(pfd_addr, bufAddr, sizeof(double*)*ADDR_BUFSIZE);
	countAddr = 0;
    }
}

void ad_tape_stop_addr()
{
    if (countAddr > 0) {
	ad_tape_write(pfd_addr, bufAddr, sizeof(double*)*countAddr);
	countAddr = 0;
    }
    fflush(pfd_addr);
}


double* ad_tape_get_addr()
{
    double* x;
    if (countAddr <= 0) {
	countAddr = ad_tape_read( pfd_addr, bufAddr, 
			  sizeof(double*)*ADDR_BUFSIZE ) / sizeof(double*);
	if (!countAddr) {
	    printf("Error: null get addr\n");
	    exit(1);
        }
    }
    x = bufAddr[--countAddr];
    /*    printf("get:%d ", x);*/
    
    if (x >= (double*)stack_low && x <= (double*)stack_high) {
	x += STACK_OFFSET;
    }
    return x;
}
    



#line 351 "ad_adj.w"
void ad_tape_open_alpha()
{
    countAlpha = 0;
    pfd_alpha = ad_tape_open(".adic.tape.alpha");
}

void ad_tape_start_alpha()
{
    countAlpha = 0;
}

void ad_tape_rev_start_alpha()
{
    countAlpha = 0;
    ad_tape_rev_start(pfd_alpha);
}

void ad_tape_put_alpha(double alpha)
{
    if (alpha == 0.0) {
	/*	printf("A:%f", alpha);*/
    }
    
    bufAlpha[countAlpha++] = alpha;
    if (countAlpha >= ALPHA_BUFSIZE) {
	ad_tape_write(pfd_alpha, bufAlpha, sizeof(double)*ALPHA_BUFSIZE);
	countAlpha = 0;
    }
}

void ad_tape_stop_alpha()
{
    if (countAlpha > 0) {
	ad_tape_write(pfd_alpha, bufAlpha, sizeof(double)*countAlpha);
	countAlpha = 0;
    }
    fflush(pfd_alpha);
}


double ad_tape_get_alpha()
{
    double alpha;
    
    if (countAlpha <= 0) {
	countAlpha = ad_tape_read( pfd_alpha, bufAlpha, 
	                           sizeof(double)*ALPHA_BUFSIZE );
	countAlpha /= sizeof(double);
    }
    alpha = bufAlpha[--countAlpha];
    return alpha;
	

}
    




#line 413 "ad_adj.w"
void ad_adj_axpy_0(double* dz)
{
    ad_tape_put_op(OP_ASSIGN_ZERO, 0);
    ad_tape_put_addr(dz);
}

void ad_adj_axpy_copy(double* dz, double* dx)
{
    int	    i;
    double  alpha, *loc;
    va_list ap;
    va_start(ap, dz);

    ad_tape_put_op(OP_ASSIGN_COPY, 0);
    ad_tape_put_addr(dx);
    ad_tape_put_addr(dz);
}

void ad_adj_axpy_1(double* dz, double alpha, double* da)
{
    int	    i;
    ad_tape_put_op(OP_ASSIGN_ONE, 0);
    ad_tape_put_alpha(alpha);
    ad_tape_put_addr(da);
    ad_tape_put_addr(dz);
}


void ad_adj_axpy_n(int arity, double* dz, ...)
{
    int	    i;
    double  alpha, *loc;
    va_list ap;
    va_start(ap, dz);

    ad_tape_put_op(OP_ASSIGN, arity);
    for (i=0; i < arity; i++) {
	alpha = va_arg(ap, double);
	ad_tape_put_alpha(alpha);
	ad_tape_put_addr(va_arg(ap, double*));
    }
    ad_tape_put_addr(dz);
    va_end(ap);
}



#line 467 "ad_adj.w"
void ad_adj_reverse()
{
    int		i, mod, op;
    double	dz, x, *z, *a, *b, *c, *d;

    ad_tape_stop_alpha();
    ad_tape_stop_op();
    ad_tape_stop_addr();
    
    ad_tape_rev_start_alpha();
    ad_tape_rev_start_op();
    ad_tape_rev_start_addr();

    nullMemory();

    while (op = ad_tape_get_op(&mod)) {
	switch (op) {
	  case OP_ASSIGN_ZERO:
	      z = ad_tape_get_addr();
	      printf("Z:%d %.3f\n", z, *z);
	      *z = 0.0;
	      
	    break;

	  case OP_ASSIGN_COPY:
            z = ad_tape_get_addr();
            printf("C:%d %.3f\n", z, *z);
	    dz = *z;		
            *z = 0.0;
	    a = ad_tape_get_addr();
	    *a += dz;
            /*printf("a:%d %.3f  ", a, *a);*/
	    break;

	  case OP_ASSIGN_ONE:
            z = ad_tape_get_addr();
            printf("1:%d %.3f\n", z, *z);
	    dz = *z;		
            *z = 0.0;
	    a = ad_tape_get_addr();
	    x = ad_tape_get_alpha();
	    *a += x * dz;
	    /*printf("a:%d %.3f A:%.3f ", a, *a, x);*/
	    break;

	  case OP_ASSIGN:
            z = ad_tape_get_addr();
            printf("%d:%d %.3f\n", mod, z, *z);
	    dz = *z;		
            *z = 0.0;
	    for (i = 0; i < mod; i++) {
		a = ad_tape_get_addr();
		x = ad_tape_get_alpha();
		
	        *a += x * dz;
		/*printf("a[%d]:%d %.3f A:%.3f ", i, a, *a, x);*/
	    }
	    break;

          default:
	    printf("Unknown tape op: %d\n", op);
	    exit(1);
        }
    }
}
	



#line 537 "ad_adj.w"
void ad_adj_enter()
{
    char    local;
    /*printf("stacktop: %d\n", &local);*/
    if (stack_low == 0 || &local < (char*)stack_low) {
	stack_low = &local;
    }
    else if (&local > (char*)stack_high) {
	stack_high = &local;
    }
}



#line 563 "ad_adj.w"
void* ad_adj_malloc(size_t size)
{
    char* p = (char*)malloc(size);
    if (!heap_low || p < heap_low) {
	heap_low = p;
    }
    if (p+size > heap_high) {
	heap_high = p;
    }
    return p;
}



#line 578 "ad_adj.w"
void* ad_adj_calloc(size_t size, size_t elsize)
{
    char* p = (char*)calloc(size, elsize);
    if (!heap_low || p < heap_low) {
	heap_low = p;
    }
    if (p+size*elsize > heap_high) {
	heap_high = p;
    }
    return p;
}



#line 593 "ad_adj.w"
void* ad_adj_realloc(void* ptr, size_t size)
{
    char* p = (char*)realloc(ptr, size);
    if (!heap_low || p < heap_low) {
	heap_low = p;
    }
    if (p+size > heap_high) {
	heap_high = p;
    }
    return p;
}


void ad_adj_free(void* p)
{
    free(p);
}





#line 625 "ad_adj.w"
void nullMemory()
{
    if (heap_low) {
        memset(heap_low, '\0', heap_high - heap_low);
    }

    if (stack_high) {
        memset( stack_low + sizeof(double)*STACK_OFFSET, 
	        '\0', stack_high - stack_low );
    }
}





#line 650 "ad_adj.w"
void ad_adj_init()
{
    char    local;
    
    ad_tape_open_op();
    ad_tape_open_addr();
    ad_tape_open_alpha();
}


void ad_adj_final()
{
}





#line 673 "ad_adj.w"
void ad_adj_set_seed(double* ploc, double d)
{
    *ploc = d;
}
    

void ad_adj_extract(double* ploc, double adj)
{
    adj = *ploc;
}


void ad_adj_var_init(void *x, int size)
{
    memset(x, '\0', size);
}
    





