
# Parse a simple assignment statment of the form "foo = bar;".  
# Return "foo" (with whitespace trimmed) for LHS, and "bar"
# with whitespace trimmed for RHS.


package AssignmentParse;

$LHS = 0;
$RHS = 1;
$OK = 2;

use PurseConstants qw($ADINTR_UC_PREFIX);

sub new {
    my $type = shift;
    my $line = shift;
    my $lhs, $rhs, $ok;

    if ( $line =~ m/^\s*(\S+)\s*=\s*(\S+)\s*;\s*$/ )
    {
	$lhs = $1;
	$rhs = $2;
	$ok = 1;
    }
    else
    {
	$lhs = "AssignmentParse: PARSE ERROR";
	$rhs = "AssignmentParse: PARSE ERROR";
	$ok = 0;

	warn "AssignmentParse unable to parse line:\n$line\n";
    }

    my $self = [ $lhs, $rhs, $ok ];
    bless $self, $type;
}

sub LHS {
    my $self = shift;
    return $self->[$LHS];
}

sub RHS {
    my $self = shift;
    return $self->[$RHS];
}

sub OK {
    my $self = shift;
    return $self->[$OK];
}

#
# Input: an uppercase intrinsic function
# Usage:
# $assign = new AssignmentParse("fx = 0.0;");
# $assign->ADIntr_Partials_assign("ACOS");
#
# returns "fx = ADIntr_Partials[ADINTR_ACOS][ADINTR_FX];\n"
#

sub ADIntr_Partials_assign {
    my $self = shift;
    my $uc_func = shift;
    my $partial, $uc_partial;

    if ( ! $self->OK() ) 
    {
	warn "Converting garbage to ADIntr_Partials assignment\n";
    }

    $partial = $self->LHS();
    $uc_partial = $partial;
    $uc_partial =~ tr/a-z/A-Z/;
    
    return "${partial} = ADIntr_Partials[${ADINTR_UC_PREFIX}${uc_func}][${ADINTR_UC_PREFIX}${uc_partial}];\n";

}
