package PurseErrors;

use Exporter;
use LineInFile;
use PurseConstants;
use BufferedInput qw(input_line_number);

@ISA = qw(Exporter);
@EXPORT = qw(die_input_problem      warn_input_problem
	     die_template_problem   warn_template_problem
	     die_template_input_problem
	     warn_template_input_problem

	     $program
	     $line_in_template
	     $current_file
	     $template_file);

=head 1 Warning and Errors

As recommended by the GNU coding standards, error messages and warning
should have the form
     PROGRAM:SOURCE-FILE-NAME:LINENO: MESSAGE

or
     PROGRAM: MESSAGE

The variable ``purse'' is used by all death messages to indicate who
is complaining. 

=cut

$program = $PurseConstants::program;

$current_file = "UNKNOWN";
$template_file = "UNKNOWN";
$line_in_template = -1;

# Provided by BufferedInput
#sub input_line_number {
#    return -1;
#}

=head 1 Interface

     die_input_problem
     warn_input_problem

     die_template_problem
     warn_template_problem

     die_template_input_problem
     warn_template_input_problem

=cut

sub die_input_problem {
    $line_in_input_file = &input_line_number;
    print STDERR "$program:$current_file:$line_in_input_file: Fatal error: $_[0]\n";
    die "\n";
}

sub warn_input_problem {
    $line_in_input_file = &input_line_number;
    print STDERR "$program:$current_file:$line_in_input_file: Warning: $_[0]\n";
}


sub die_template_problem {
    if ( defined( $_[1] ) ) {
	$line_in_template = $_[1]->LineNumber();
    }
    print STDERR "$program:$template_file:$line_in_template: Fatal error: $_[0]\n";
    die "\n";
}

sub warn_template_problem {
    if ( defined($_[1] ) ) {
	$line_in_template = $_[1] -> LineNumber();
    }
    print STDERR "$program:$template_file:$line_in_template: Warning: $_[0]\n";
}


sub warn_template_input_problem {
    if ( defined($_[1] ) ) {
	$line_in_template = $_[1] -> LineNumber();
    }
    $line_in_input_file = &input_line_number;
    print STDERR "$program:$current_file:$line_in_input_file: Warning expanding template $template_file, line $line_in_template: $_[0]\n";
}

sub die_template_input_problem {
    if ( defined($_[1] ) ) {
	$line_in_template = $_[1] -> LineNumber();
    }
    $line_in_input_file = &input_line_number;
    print STDERR "$program:$current_file:$line_in_input_file: Fatal error expanding template $template_file, line $line_in_template: $_[0]\n";
    die "\n";
}
