#use English;
use Exporter;
#use strict; # ha ha

package ExpandTemplate;
@ISA = qw(Exporter);
@EXPORT = (expand_template);
@EXPORT_OK = ();

use PurseConstants qw($FILE_NUMBER_VAR_NAME $ADINTR_UC_PREFIX);
use PurseErrors;
use TemplateFile;
use LineInFile;
#use Carp; # debugger did not like bare croak
use AssignmentParse;

    $x="ERROR_X"; $y = "ERROR_Y"; $z = "ERROR_Z"; 
    $fx = "ERROR_FX"; $fy = "ERROR_Fy"; 
    $fxx = "ERROR_FXX"; $fxy = "ERROR_FXY"; $fyy = "ERROR_FYY";
# $line_in_template = -1; # global from PurseErrors
 @template_text = ();
 $last_gotten = undef;
 $expansion_type = "NONE AT ALL";
 $mode = "NONE";
sub get_next_template_line 
{
    my ($out_ref) = @_;
    $last_gotten = shift (@template_text);

    if ( defined($last_gotten) ) {
        $line_in_template = $last_gotten->LineNumber();
        $$out_ref = $last_gotten->Text();
        return 1;
    } else {
        $$out_ref = undef;
        return 0;
    }
}
sub unget_template_line {
    push (@template_text, $last_gotten);
}
sub make_exception_handler_call
{
    my ($tf, $mode, $expansion_type) = @_;

    my $function = $tf->FunctionName();
    $function =~ tr/A-Z/a-z/;

    my $uc_func = $function;
    $uc_func =~ tr/a-z/A-Z/;

    my $deriv_order;

    if ( $expansion_type eq "FVAL" ) {
        $deriv_order = 0;
        &die_template_input_problem("Exception handler call generated for FVAL expansion");
    } elsif ( $expansion_type eq "FIRST" ) {
        $deriv_order = 1;
    } elsif ( $expansion_type eq "SECOND" ) {
        $deriv_order = 2 ;
    } else {
        die "${PurseConstants::program}: Internal Error: Bad value `$expansion_type' for expansion type\n";
    }

    my $call = "";
    if ( $mode ne "P" ) {
    $call = "adintr_${function}($deriv_order, $FILE_NUMBER_VAR_NAME, __LINE__";

    if ($deriv_order >= 1) {
        $call .= ", &fx";
        if ( $tf->NumberOfArguments() >= 2 ) {
            $call .= ", &fy";
        }
    } 
    if ($deriv_order >= 2) {
        $call .= ", &fxx";
        if ( $tf->NumberOfArguments() >= 2 ) {
            $call .= ", &fxy, &fyy";
        }
    }
    $call .= ");";
    } else {
{
    my @to_insert;
    my ($line, $line_info);
    
    if ($tf->HasExceptionalCode()) {
        @to_insert = $tf->ExceptionalCode();
    for $line_info  (@to_insert) {
        $line = $line_info->Text();
        next if $line =~ m/^\s*$/;

        # Skip derivatives of too high order
        next if ( ($deriv_order == 1)
                  && ($line =~ m/\bf[xy][xy]\b/ ));
        $call .= "$line\n";
    }
    } else {
        @to_insert = $tf->ExceptionalValues();
for $line_info (@to_insert)
{
    $line = $line_info->Text();
    next if $line =~ m/^\s*$/;

    # Skip derivatives of too high order
    next if ( ($deriv_order == 1)
             && ($line =~ m/\bf[xy][xy]\b/ ));

    my $assign, $partial, $uc_partial;

    $assign = new AssignmentParse($line);

    &die_template_problem("Unable to parse exceptional values line",$line_info)
        if ! $assign->OK();

    $call .= $assign->ADIntr_Partials_assign($uc_func);
}
    }
}
    }

    $call =~ s/\n$//;
    return $call;
}
$exception_handler_regex = "EXCEPTION_HANDLER\s*\(\s*\)";
$preceding_was_already_true = 1;
# These are constants
$COMMENT_TOK = -1;
$IFDEF_TOK = 1;
$IFNDEF_TOK = 2;
$ELIF_TOK = 3;
$ELSE_TOK = 4;
$ENDIF_TOK = 5;
$FINISH_TOK = 6;

%get_token_number = ("ifdef", $IFDEF_TOK,
                     "ifndef", $IFNDEF_TOK,
                     "elif", $ELIF_TOK,
                     "else", $ELSE_TOK,
                     "endif", $ENDIF_TOK,
                     "finish", $FINISH_TOK);

$directive_line_begin_regex = '\#\s*';
$directive_comment_regex = '\#\#\s*';

$recognized_hash_directives_regex = '(ifdef|ifndef|else|elif|endif|finish)';

%directive_tests_state = ($IFDEF_TOK, 1,
                          $IFNDEF_TOK, 1,
                          $ELIF_TOK, 1,
                          $ELSE_TOK, 0,
                          $ENDIF_TOK, 0,
                          $FINISH_TOK, 1 );

$testable_states_regex = '(PERFORMANCE|FVAL|FIRST|SECOND)';

$type_regex = '(\b)TYPE\s*\(([^\),]+)(,([^\)]+))?\)';
#                           ^^^^^^^   ^^^^^^^^       
#            argument:       first     second
sub snarf_rest_of_ifdef {
    local ($line);
    local ($found_match) = 0;
    local ($directive, $test_condition);

  scanner_loop:
    while (&get_next_template_line(\$line))
    {
        if ( $line =~ m/$directive_line_begin_regex/o )
        {
            # Safety for the users -- don't want to skip over syntax errors
            # or other bad stuff.
            ($directive, $test_condition) = &syntax_check_line($line);

            if ( $directive == $ENDIF_TOK )
            {
                $found_match = 1;
                last scanner_loop;
            }
            
            # See if directive begins a nested "if..." structure.
            if ( $directive == $IFDEF_TOK
                || $directive == $IFNDEF_TOK )
            {
                &snarf_rest_of_ifdef();
            }
        }
    }

    &die_template_problem("Unterminated #ifdef in template file\n")
        if ( ! $found_match );
}
sub snarf_until_next_branch {
    local ($line);
    local ($found_match) = 0;
    local ($directive, $test_condition);

  scanner_loop:
    while (&get_next_template_line(\$line))
    {
        if ( $line =~ m/$directive_line_begin_regex/o )
        {
            # Safety for the users -- don't want to skip over syntax errors
            # or other bad stuff.
            ($directive, $test_condition) = &syntax_check_line($line);

            if ( $directive == $ELSE_TOK
                || $directive == $ELIF_TOK 
                || $directive == $ENDIF_TOK )
            {
                $found_match = 1;

                # Put it back to be analyzed again in the proper context.
                &unget_template_line ($line);

                # There are two ways to arrive at an else/elif/endif:
                # either you can skip the preceding section or you can
                # go through the preceding section & bumble into them.
                # If the latter, *they* should be skipped, so we need
                # a way of indicating what happened.

                if ( $directive != $ENDIF_TOK) {
                    $preceding_was_already_true = 0;
                }

                last scanner_loop;
            }

            # See if directive begins a nested "if..." structure.
            if ( $directive == $IFDEF_TOK
                || $directive == $IFNDEF_TOK )
            {
                &snarf_rest_of_ifdef();
            }
        }
    }

    &die_template_problem("Unterminated #ifdef in template file\n")
        if ( ! $found_match );
}
sub syntax_check_line {
    local ($_) = @_;
#    local ($original_line) = $_;
    local ($directive, $state_tested);
    
    # We just approve lines that are not directives.
    return undef if ! m/${directive_line_begin_regex}/o;

    # Comments are always OK.
    return ($COMMENT_TOK,undef) if m/${directive_comment_regex}/o;

    # Check for recognized directive at start.
    #    Die if not recognized
    #    Remember directive & remove start of line if we do recognize it.
    #
    if ( ! s/${directive_line_begin_regex}${recognized_hash_directives_regex}//o )
    {
        &die_template_problem ("Unrecognized #-directive in template line");
    }
    else 
    {
        $directive = $1;        # Just in case $1 doesn't last too long.
        $directive = $get_token_number{"$directive"};
    }

    # Now make sure there is a state to test, or not, whichever is
    # required by the syntax of the directive we saw.

    if ( $directive_tests_state{$directive} )
    {
        if ( ! m/\s*${testable_states_regex}\s*$/o )
        {
            &die_template_problem("Test condition not recognized");
        }
        $state_tested = $1;
    }
    else # (does not require state)
    {
        # Allow /* ... */ comments after else, endif (non-nesting, of course)
        if ( ! m/^\s*(|\/\*.*\*\/)\s*$/ ) 
        {
            &die_template_problem("Garbage after end of directive");
        }
    }
    return ($directive, $state_tested);
}
sub handle_TYPE_macro {
    local($boundary_char, $real, $imaginary) = @_;

    $real_expansion = &handle_one_TYPE($real);

    if ( $precision eq "C" || $precision eq "Z" )
    {
        if ( ! defined ($imaginary) ) {
            $imaginary = "0.0";
        }
        $imaginary_expansion = &handle_one_TYPE($imaginary);

        return "${boundary_char}(${real_expansion},${imaginary_expansion})";
    }
    else
    {
        return "${boundary_char}${real_expansion}";
    }
}
sub handle_one_TYPE {
    local ($constant) = @_;
    
    # Trim whitespace
    $constant =~ s/^\s*//; $constant =~ s/\s*$//;

    local($has_exponent) = ($constant =~ m/ed/i );

    if ( ! $has_exponent )
    {
        # Default to double precision
        $constant .= "d0";
    }

    if ( $precision eq "S" || $precision eq "C") {
        $constant =~ s/d/e/gi;
    } elsif ( $precision eq "D" || $precision eq "Z" ) {
        $constant =~ s/e/d/gi;
    }

    return $constant;
}
sub check_condition {
    local ($condition) = @_;
    local ($result);

    if ( $condition eq "PERFORMANCE" ) {
        $result = ( $mode eq "P" );
    }
    else {
        $result = ( $condition eq $expansion_type );
    }
    return $result;
} 
sub expand_template {
    my ($tf,$precision,$expansion_type,$mode, @arg_names) = @_;
    my @line = ();

    $z = shift(@arg_names);
    $x = shift(@arg_names);
    $y = shift(@arg_names);
    $fx = shift(@arg_names);
    $fy = shift(@arg_names);
    $fxx = shift(@arg_names);
    $fxy = shift(@arg_names);
    $fyy = shift(@arg_names);

    $ExpandTemplate::expansion_type = $expansion_type;
    $ExpandTemplate::mode = $mode;

    @ExpandTemplate::template_text = $tf->TemplateCode();

  template_processing_loop:
    # Loop as long as we have things left in the list (and keep going
    # when we get blank lines).
    while (&get_next_template_line(\$_))
    {
{
if ( m/${directive_line_begin_regex}/o )
{
    local ($directive, $test_condition) = &syntax_check_line ($_);

    if ($directive == $COMMENT_TOK) {
        # Nothing to do
    }
    elsif ( $directive == $FINISH_TOK) {
if ( &check_condition($test_condition) ) 
{
    # Bail out
    last template_processing_loop;
}
    }
    elsif ( $directive == $ENDIF_TOK ) {
        # Back to emitting; since we emit by default, this is a no-op.
    }
    elsif ( $directive == $ELSE_TOK ) {
# See web doc.
if ( $preceding_was_already_true ) {
    &snarf_rest_of_ifdef;
}
else
{
    $preceding_was_already_true = 1;
}
    } 
    elsif ( $directive == $ELIF_TOK ) {
        if ( $preceding_was_already_true ) {
# See web doc.
if ( $preceding_was_already_true ) {
    &snarf_rest_of_ifdef;
}
else
{
    $preceding_was_already_true = 1;
}
        }
        else
        {
            $preceding_was_already_true = 1;
{
    local ($emit);

    $emit = &check_condition($test_condition);

    if ( $directive == $IFNDEF_TOK ) { $emit = ! $emit; } 

    if ( ! $emit ) { &snarf_until_next_branch; }
}
        }
    }
    elsif ($directive == $IFDEF_TOK 
           || $directive == $IFNDEF_TOK )
    {
{
    local ($emit);

    $emit = &check_condition($test_condition);

    if ( $directive == $IFNDEF_TOK ) { $emit = ! $emit; } 

    if ( ! $emit ) { &snarf_until_next_branch; }
}
    }
    else {
        die "$purse: Internal error: unrecognized directive code $directive. Stopped ";
    }    
    next;
}
# Skip second order terms if doing first-order expansion.
if ( $expansion_type eq "FIRST" )
{
    next if ( m/\bf[xy][xy]\b/ );
}

# Skip first and second order terms if doing function-value expansion
if ( $expansion_type eq "FVAL" )
{
    next if m/\bf[xy](|[x-y])\b/;
}

    # Make exception handler call if that's what we see.
    if ( m/$exception_handler_regex/io )
    {
        $_ = &make_exception_handler_call ($tf, $mode, $expansion_type);
    }


# Bracket each substitution var with \000 ... \000
s/(\b)(z|x|y|fx|fy|fxx|fxy|fyy)(\b)/$1\000$2\000$3/ig;

# Appropriately rename
s/\000(z)\000/$z/ig;               # Function Result

s/\000(x)\000/$x/ig;               # Function Input(s)
s/\000(y)\000/$y/ig;

s/\000(fx)\000/$fx/ig;             # First order partials
s/\000(fy)\000/$fy/ig;

s/\000(fxx)\000/$fxx/ig;           # Second order partials
s/\000(fxy)\000/$fxy/ig;
s/\000(fyy)\000/$fyy/ig;


#    < < Re-indent code > >

    # Output the line (if we get here, it should go out the door).
    push(@line, $_);
}
    }

    # Remove leading and trailing blank lines.
    shift(@line) while $line[0] =~ m/^\s*$/;
    pop(@line) while $line[$#line] =~ m/^\s*$/;

    # Stripping off these opening and closing braces solves the problem of 
    # "#finish" directives unbalancing them.
    
    # Remove leading open brace, trailing closing brace, if it exists.
    shift (@line) if $line[0] =~ m/^\s*\{\s*$/;
    pop (@line) if $line[$#line] =~ m/^\s*\}\s*$/;

    # Put in informative comments
    if ( $PurseConstants::INCLUDE_COMMENTS_IN_OUTPUT ) 
    {
        unshift(@line,"/* Start Template: " . $tf->FunctionName() . " */");
        push(@line,"/* End Template: " . $tf->FunctionName() . " */");
    }

    return (@line);
}

1;
