extern int hess_size_1, hess_size_2;
extern int hess_maxsize_1, hess_maxsize_2;
  


void 
ad_fh_faddad (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b, double *g_b, double *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i] + g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i] + h_b[i];
      }
  }
}


void 
ad_fh_faddbd (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i];
      }
  }
}


void 
ad_fh_faddcd (double c, double *g_c, double *h_c, double a, double b, double *g_b, double *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_b[i];
      }
  }
}


void 
ad_fh_fsubad (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b, double *g_b, double *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i] - g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i] - h_b[i];
      }
  }
}


void 
ad_fh_fsubbd (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i];
      }
  }
}


void 
ad_fh_fsubcd (double c, double *g_c, double *h_c, double a, double b, double *g_b, double *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = -g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = -h_b[i];
      }
  }
}


void 
ad_fh_fmulad (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b, double *g_b, double *h_b)
{

  int n, j, i;

  {
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = b * h_a[n] + a * h_b[n] + g_a[i] * g_b[j] + g_a[j] * g_b[i];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = b * g_a[i] + a * g_b[i];
      }
  }
}


void 
ad_fh_fmulbd (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = b * g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = b * h_a[i];
      }
  }
}


void 
ad_fh_fmulcd (double c, double *g_c, double *h_c, double a, double b, double *g_b, double *h_b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = a * g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = a * h_b[i];
      }
  }
}


void 
ad_fh_fdivad (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b, double *g_b, double *h_b)
{

  int n, j, i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = temp3 * (h_a[n] - temp1 * h_b[n] + temp2 * g_b[i] * g_b[j] - temp3 * (g_a[i] * g_b[j] + g_a[j] * g_b[i]));
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = temp3 * g_a[i] - temp1 * temp3 * g_b[i];
      }
  }
}


void 
ad_fh_fdivbd (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i] / b;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i] / b;
      }
  }
}


void 
ad_fh_fdivcd (double c, double *g_c, double *h_c, double a, double b, double *g_b, double *h_b)
{

  int n, j, i;
  double temp1, temp2, temp3;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = temp1 * h_b[n] - temp2 * g_b[i] * g_b[j];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = temp1 * g_b[i];
      }
  }
}


void 
ad_fh_fcopyd (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i];
      }
  }
}


void 
ad_fh_fmind (double c, double *g_c, double *h_c, double a, double *g_a, double *h_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = -g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = -h_a[i];
      }
  }
}


void 
ad_fh_fzerod (double c, double *g_c, double *h_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = 0.0;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = 0.0;
      }
  }
}


void 
ad_fh_faddas (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b, float *g_b, float *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i] + g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i] + h_b[i];
      }
  }
}


void 
ad_fh_fintrin1d (double fa, double faa, double c, double *g_c, double *h_c, double a, double *g_a, double *h_a)
{

  int n, j, i;
  double temp;

  {
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	temp = faa * g_a[i];
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = fa * h_a[n] + temp * g_a[j];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = fa * g_a[i];
      }
  }
}


void 
ad_fh_fintrin2d (double fa, double fb, double faa, double fab, double fbb, double c, double *g_c, double *h_c, double a, double *g_a, double *h_a, double b, double *g_b, double *h_b)
{

  int n, j, i;
  double temp1, temp2;

  {
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	temp1 = faa * g_a[i] + fab * g_b[i];
	temp2 = fab * g_a[i] + fbb * g_b[i];
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = fa * h_a[n] + fb * h_b[n] + temp1 * g_a[j] + temp2 * g_b[j];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = fa * g_a[i] + fb * g_b[i];
      }
  }
}


void 
ad_fh_faddbs (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i];
      }
  }
}


void 
ad_fh_faddcs (float c, float *g_c, float *h_c, float a, float b, float *g_b, float *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_b[i];
      }
  }
}


void 
ad_fh_fsubas (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b, float *g_b, float *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i] - g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i] - h_b[i];
      }
  }
}


void 
ad_fh_fsubbs (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i];
      }
  }
}


void 
ad_fh_fsubcs (float c, float *g_c, float *h_c, float a, float b, float *g_b, float *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = -g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = -h_b[i];
      }
  }
}


void 
ad_fh_fmulas (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b, float *g_b, float *h_b)
{

  int n, j, i;

  {
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = b * h_a[n] + a * h_b[n] + g_a[i] * g_b[j] + g_a[j] * g_b[i];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = b * g_a[i] + a * g_b[i];
      }
  }
}


void 
ad_fh_fmulbs (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = b * g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = b * h_a[i];
      }
  }
}


void 
ad_fh_fmulcs (float c, float *g_c, float *h_c, float a, float b, float *g_b, float *h_b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = a * g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = a * h_b[i];
      }
  }
}


void 
ad_fh_fdivas (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b, float *g_b, float *h_b)
{

  int n, j, i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = temp3 * (h_a[n] - temp1 * h_b[n] + temp2 * g_b[i] * g_b[j] - temp3 * (g_a[i] * g_b[j] + g_a[j] * g_b[i]));
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = temp3 * g_a[i] - temp1 * temp3 * g_b[i];
      }
  }
}


void 
ad_fh_fdivbs (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i] / b;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i] / b;
      }
  }
}


void 
ad_fh_fdivcs (float c, float *g_c, float *h_c, float a, float b, float *g_b, float *h_b)
{

  int n, j, i;
  float temp1, temp2, temp3;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = temp1 * h_b[n] - temp2 * g_b[i] * g_b[j];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = temp1 * g_b[i];
      }
  }
}


void 
ad_fh_fcopys (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = h_a[i];
      }
  }
}


void 
ad_fh_fmins (float c, float *g_c, float *h_c, float a, float *g_a, float *h_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = -g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = -h_a[i];
      }
  }
}


void 
ad_fh_fzeros (float c, float *g_c, float *h_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = 0.0;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = 0.0;
      }
  }
}


void 
ad_fh_fintrin1s (float fa, float faa, float c, float *g_c, float *h_c, float a, float *g_a, float *h_a)
{

  int n, j, i;
  float temp;

  {
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	temp = faa * g_a[i];
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = fa * h_a[n] + temp * g_a[j];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = fa * g_a[i];
      }
  }
}


void 
ad_fh_fintrin2s (float fa, float fb, float faa, float fab, float fbb, float c, float *g_c, float *h_c, float a, float *g_a, float *h_a, float b, float *g_b, float *h_b)
{

  int n, j, i;
  float temp1, temp2;

  {
    n = 0;
    for (i = 0; i < hess_size_1; i++)
      {
	temp1 = faa * g_a[i] + fab * g_b[i];
	temp2 = fab * g_a[i] + fbb * g_b[i];
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    h_c[n] = fa * h_a[n] + fb * h_b[n] + temp1 * g_a[j] + temp2 * g_b[j];
	    n = n + 1;
	  }
      }
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = fa * g_a[i] + fb * g_b[i];
      }
  }
}


void 
ad_fh_accumg1d (double lg1, double *g_c, double *g_a)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = lg1 * g_a[i];
      }
  }
}


void 
ad_fh_accumg2d (double lg1, double lg2, double *g_c, double *g_a, double *g_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = lg1 * g_a[i] + lg2 * g_b[i];
      }
  }
}


void 
ad_fh_accumg3d (double lg1, double lg2, double lg3, double *g_d, double *g_a, double *g_b, double *g_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_d[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i];
      }
  }
}


void 
ad_fh_accumg4d (double lg1, double lg2, double lg3, double lg4, double *g_lhs, double *g_a, double *g_b, double *g_c, double *g_d)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_lhs[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i];
      }
  }
}


void 
ad_fh_accumg5d (double lg1, double lg2, double lg3, double lg4, double lg5, double *g_lhs, double *g_a, double *g_b, double *g_c, double *g_d, double *g_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_lhs[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i] + lg5 * g_e[i];
      }
  }
}


void 
ad_fh_accumhg1d (double lg1, double *h_c, double *h_a)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = lg1 * h_a[i];
      }
  }
}


void 
ad_fh_accumhg2d (double lg1, double lg2, double *h_c, double *h_a, double *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = lg1 * h_a[i] + lg2 * h_b[i];
      }
  }
}


void 
ad_fh_accumhg3d (double lg1, double lg2, double lg3, double *h_d, double *h_a, double *h_b, double *h_c)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_d[i] = lg1 * h_a[i] + lg2 * h_b[i] + lg3 * h_c[i];
      }
  }
}


void 
ad_fh_accumhg4d (double lg1, double lg2, double lg3, double lg4, double *h_lhs, double *h_a, double *h_b, double *h_c, double *h_d)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_lhs[i] = lg1 * h_a[i] + lg2 * h_b[i] + lg3 * h_c[i] + lg4 * h_d[i];
      }
  }
}


void 
ad_fh_accumhg5d (double lg1, double lg2, double lg3, double lg4, double lg5, double *h_lhs, double *h_a, double *h_b, double *h_c, double *h_d, double *h_e)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_lhs[i] = lg1 * h_a[i] + lg2 * h_b[i] + lg3 * h_c[i] + lg4 * h_d[i] + lg5 * h_e[i];
      }
  }
}


void 
ad_fh_accumh1d (double lh1, double *h_v0, double *g_v1)
{

  int n, j, i;
  double temp1;

  {
    if (lh1 != 0.0)
      {
	n = 0;
	for (i = 0; i < hess_size_1; i++)
	  {
	    temp1 = g_v1[i] * lh1;
	    for (j = 0; j < i + (1 - 0); j++)
	      {
		h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		n = n + 1;
	      }
	  }
      }
  }
}


void 
ad_fh_accumh2d (double lh1, double lh2, double *h_v0, double *g_v1, double *g_v2)
{

  int iflag, n, j, i;
  double temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      temp2 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh3d (double lh1, double lh2, double lh3, double *h_v0, double *g_v1, double *g_v2)
{

  int iflag, n, j, i;
  double temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      temp2 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 4:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh3;
	      temp2 = g_v2[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j] + temp2 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 5:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 6:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2 + g_v1[i] * lh3;
	      temp2 = g_v2[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j] + temp2 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 7:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh3;
	      temp2 = g_v2[i] * lh2 + g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh4d (double lh1, double *h_v0, double *g_v1, double *g_v2)
{

  int n, j, i;
  double temp1, temp2;

  {
    if (lh1 != 0.0)
      {
	n = 0;
	for (i = 0; i < hess_size_1; i++)
	  {
	    temp1 = g_v2[i] * lh1;
	    temp2 = g_v1[i] * lh1;
	    for (j = 0; j < i + (1 - 0); j++)
	      {
		h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		n = n + 1;
	      }
	  }
      }
  }
}


void 
ad_fh_accumh5d (double lh1, double lh2, double *h_v0, double *g_v1, double *g_v2, double *g_v3)
{

  int iflag, n, j, i;
  double temp1, temp2, temp3;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1;
	      temp2 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh6d (double lh1, double lh2, double lh3, double *h_v0, double *g_v1, double *g_v2, double *g_v3, double *g_v4)
{

  int iflag, n, j, i;
  double temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1;
	      temp2 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 4:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      case 5:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      case 6:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh2 + g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh2;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j] + temp3 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      case 7:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v3[i] * lh2 + g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh2;
	      temp4 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j] + temp4 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh7d (double lh1, double lh2, double lh3, double *h_v0, double *g_v1, double *g_v2, double *g_v3)
{

  int iflag, n, j, i;
  double temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 4:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 5:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 6:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2 + g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh2;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 7:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh2 + g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh2;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh8d (double lh1, double lh2, double *h_v0, double *g_v1, double *g_v2)
{

  int iflag, n, j, i;
  double temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_taccumg1d (double lg1, void *t_lhs, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg2d (double lg1, double lg2, void *t_lhs, void *t_a, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg3d (double lg1, double lg2, double lg3, void *t_lhs, void *t_a, void *t_b, void *t_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0) * hess_maxsize_1 + i] + lg3 * ((double *) t_c)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg4d (double lg1, double lg2, double lg3, double lg4, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0) * hess_maxsize_1 + i] + lg3 * ((double *) t_c)[(0) * hess_maxsize_1 + i] + lg4 * ((double *) t_d)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg5d (double lg1, double lg2, double lg3, double lg4, double lg5, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d, void *t_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0) * hess_maxsize_1 + i] + lg3 * ((double *) t_c)[(0) * hess_maxsize_1 + i] + lg4 * ((double *) t_d)[(0) * hess_maxsize_1 + i] + lg5 * ((double *) t_e)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg1d (double lg1, void *t_lhs, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg2d (double lg1, double lg2, void *t_lhs, void *t_a, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg3d (double lg1, double lg2, double lg3, void *t_lhs, void *t_a, void *t_b, void *t_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i] + lg3 * ((double *) t_c)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg4d (double lg1, double lg2, double lg3, double lg4, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i] + lg3 * ((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] + lg4 * ((double *) t_d)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg5d (double lg1, double lg2, double lg3, double lg4, double lg5, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d, void *t_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i] + lg3 * ((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] + lg4 * ((double *) t_d)[(0 + 1) * hess_maxsize_1 + i] + lg5 * ((double *) t_e)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumh1d (double lh1, void *t_v0, void *t_v1)
{

  int j, i;
  double temp1;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_1; i++)
	  {
	    ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	  }
      }
  }
}


void 
ad_fh_taccumh2d (double lh1, double lh2, void *t_v0, void *t_v1, void *t_v2)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + 0.5 * lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }

}


void 
ad_fh_taccumh3d (double lh1, double lh2, double lh3, void *t_v0, void *t_v1, void *t_v2)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + 0.5 * lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh3 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v2)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v1)[(0) * hess_maxsize_1 + i]) * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + 0.5 * lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh4d (double lh1, void *t_v0, void *t_v1, void *t_v2)
{

  int n, j, i;
  double temp1, temp2;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_1; i++)
	  {
	    ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	  }
      }
  }
}


void 
ad_fh_taccumh5d (double lh1, double lh2, void *t_v0, void *t_v1, void *t_v2, void *t_v3)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v3)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh2 * ((double *) t_v3)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh6d (double lh1, double lh2, double lh3, void *t_v0, void *t_v1, void *t_v2, void *t_v3, void *t_v4)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v3)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh2 * ((double *) t_v3)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh3 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v4)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v4)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh2 * ((double *) t_v3)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v4)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh2 * ((double *) t_v3)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v4)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh7d (double lh1, double lh2, double lh3, void *t_v0, void *t_v1, void *t_v2, void *t_v3)
{

  int iflag, i;
  double temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh3 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v3)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v3)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v3)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((double *) t_v3)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh8d (double lh1, double lh2, void *t_v0, void *t_v1, void *t_v2)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] * ((double *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((double *) t_v1)[(0) * hess_maxsize_1 + i] + lh2 * ((double *) t_v2)[(0) * hess_maxsize_1 + i]) * ((double *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_accumg1s (float lg1, float *g_c, float *g_a)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = lg1 * g_a[i];
      }
  }
}


void 
ad_fh_accumg2s (float lg1, float lg2, float *g_c, float *g_a, float *g_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_c[i] = lg1 * g_a[i] + lg2 * g_b[i];
      }
  }
}


void 
ad_fh_accumg3s (float lg1, float lg2, float lg3, float *g_d, float *g_a, float *g_b, float *g_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_d[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i];
      }
  }
}


void 
ad_fh_accumg4s (float lg1, float lg2, float lg3, float lg4, float *g_lhs, float *g_a, float *g_b, float *g_c, float *g_d)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_lhs[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i];
      }
  }
}


void 
ad_fh_accumg5s (float lg1, float lg2, float lg3, float lg4, float lg5, float *g_lhs, float *g_a, float *g_b, float *g_c, float *g_d, float *g_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	g_lhs[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i] + lg5 * g_e[i];
      }
  }
}


void 
ad_fh_accumhg1s (float lg1, float *h_c, float *h_a)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = lg1 * h_a[i];
      }
  }
}


void 
ad_fh_accumhg2s (float lg1, float lg2, float *h_c, float *h_a, float *h_b)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_c[i] = lg1 * h_a[i] + lg2 * h_b[i];
      }
  }
}


void 
ad_fh_accumhg3s (float lg1, float lg2, float lg3, float *h_d, float *h_a, float *h_b, float *h_c)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_d[i] = lg1 * h_a[i] + lg2 * h_b[i] + lg3 * h_c[i];
      }
  }
}


void 
ad_fh_accumhg4s (float lg1, float lg2, float lg3, float lg4, float *h_lhs, float *h_a, float *h_b, float *h_c, float *h_d)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_lhs[i] = lg1 * h_a[i] + lg2 * h_b[i] + lg3 * h_c[i] + lg4 * h_d[i];
      }
  }
}


void 
ad_fh_accumhg5s (float lg1, float lg2, float lg3, float lg4, float lg5, float *h_lhs, float *h_a, float *h_b, float *h_c, float *h_d, float *h_e)
{

  int i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	h_lhs[i] = lg1 * h_a[i] + lg2 * h_b[i] + lg3 * h_c[i] + lg4 * h_d[i] + lg5 * h_e[i];
      }
  }
}


void 
ad_fh_accumh1s (float lh1, float *h_v0, float *g_v1)
{

  int n, j, i;
  float temp1;

  {
    if (lh1 != 0.0)
      {
	n = 0;
	for (i = 0; i < hess_size_1; i++)
	  {
	    temp1 = g_v1[i] * lh1;
	    for (j = 0; j < i + (1 - 0); j++)
	      {
		h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		n = n + 1;
	      }
	  }
      }
  }
}


void 
ad_fh_accumh2s (float lh1, float lh2, float *h_v0, float *g_v1, float *g_v2)
{

  int iflag, n, j, i;
  float temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      temp2 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh3s (float lh1, float lh2, float lh3, float *h_v0, float *g_v1, float *g_v2)
{

  int iflag, n, j, i;
  float temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      temp2 = g_v2[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 4:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh3;
	      temp2 = g_v2[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j] + temp2 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 5:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 6:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2 + g_v1[i] * lh3;
	      temp2 = g_v2[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v2[j] + temp2 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 7:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh3;
	      temp2 = g_v2[i] * lh2 + g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh4s (float lh1, float *h_v0, float *g_v1, float *g_v2)
{

  int n, j, i;
  float temp1, temp2;

  {
    if (lh1 != 0.0)
      {
	n = 0;
	for (i = 0; i < hess_size_1; i++)
	  {
	    temp1 = g_v2[i] * lh1;
	    temp2 = g_v1[i] * lh1;
	    for (j = 0; j < i + (1 - 0); j++)
	      {
		h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		n = n + 1;
	      }
	  }
      }
  }
}


void 
ad_fh_accumh5s (float lh1, float lh2, float *h_v0, float *g_v1, float *g_v2, float *g_v3)
{

  int iflag, n, j, i;
  float temp1, temp2, temp3;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1;
	      temp2 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh6s (float lh1, float lh2, float lh3, float *h_v0, float *g_v1, float *g_v2, float *g_v3, float *g_v4)
{

  int iflag, n, j, i;
  float temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1;
	      temp2 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v3[i] * lh2;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 4:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      case 5:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      case 6:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh2 + g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh2;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j] + temp3 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      case 7:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh1 + g_v3[i] * lh2 + g_v4[i] * lh3;
	      temp2 = g_v1[i] * lh1;
	      temp3 = g_v1[i] * lh2;
	      temp4 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j] + temp4 * g_v4[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh7s (float lh1, float lh2, float lh3, float *h_v0, float *g_v1, float *g_v2, float *g_v3)
{

  int iflag, n, j, i;
  float temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 4:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 5:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 6:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2 + g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh2;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      case 7:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh2 + g_v3[i] * lh3;
	      temp2 = g_v1[i] * lh2;
	      temp3 = g_v1[i] * lh3;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_accumh8s (float lh1, float lh2, float *h_v0, float *g_v1, float *g_v2)
{

  int iflag, n, j, i;
  float temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j];
		  n = n + 1;
		}
	    }
	}
      case 2:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      case 3:
	{
	  n = 0;
	  for (i = 0; i < hess_size_1; i++)
	    {
	      temp1 = g_v1[i] * lh1 + g_v2[i] * lh2;
	      temp2 = g_v1[i] * lh2;
	      for (j = 0; j < i + (1 - 0); j++)
		{
		  h_v0[n] = h_v0[n] + temp1 * g_v1[j] + temp2 * g_v2[j];
		  n = n + 1;
		}
	    }
	}
      }
  }
}


void 
ad_fh_taccumg1s (float lg1, void *t_lhs, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg2s (float lg1, float lg2, void *t_lhs, void *t_a, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg3s (float lg1, float lg2, float lg3, void *t_lhs, void *t_a, void *t_b, void *t_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0) * hess_maxsize_1 + i] + lg3 * ((float *) t_c)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg4s (float lg1, float lg2, float lg3, float lg4, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0) * hess_maxsize_1 + i] + lg3 * ((float *) t_c)[(0) * hess_maxsize_1 + i] + lg4 * ((float *) t_d)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumg5s (float lg1, float lg2, float lg3, float lg4, float lg5, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d, void *t_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0) * hess_maxsize_1 + i] + lg3 * ((float *) t_c)[(0) * hess_maxsize_1 + i] + lg4 * ((float *) t_d)[(0) * hess_maxsize_1 + i] + lg5 * ((float *) t_e)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg1s (float lg1, void *t_lhs, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg2s (float lg1, float lg2, void *t_lhs, void *t_a, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg3s (float lg1, float lg2, float lg3, void *t_lhs, void *t_a, void *t_b, void *t_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i] + lg3 * ((float *) t_c)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg4s (float lg1, float lg2, float lg3, float lg4, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i] + lg3 * ((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] + lg4 * ((float *) t_d)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumhg5s (float lg1, float lg2, float lg3, float lg4, float lg5, void *t_lhs, void *t_a, void *t_b, void *t_c, void *t_d, void *t_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_lhs)[(0 + 1) * hess_maxsize_1 + i] = lg1 * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + lg2 * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i] + lg3 * ((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] + lg4 * ((float *) t_d)[(0 + 1) * hess_maxsize_1 + i] + lg5 * ((float *) t_e)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taccumh1s (float lh1, void *t_v0, void *t_v1)
{

  int j, i;
  float temp1;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_1; i++)
	  {
	    ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	  }
      }
  }
}


void 
ad_fh_taccumh2s (float lh1, float lh2, void *t_v0, void *t_v1, void *t_v2)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + 0.5 * lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }

}


void 
ad_fh_taccumh3s (float lh1, float lh2, float lh3, void *t_v0, void *t_v1, void *t_v2)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + 0.5 * lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh3 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v2)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v1)[(0) * hess_maxsize_1 + i]) * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + 0.5 * lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh4s (float lh1, void *t_v0, void *t_v1, void *t_v2)
{

  int n, j, i;
  float temp1, temp2;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_1; i++)
	  {
	    ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	  }
      }
  }
}


void 
ad_fh_taccumh5s (float lh1, float lh2, void *t_v0, void *t_v1, void *t_v2, void *t_v3)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v3)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh2 * ((float *) t_v3)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh6s (float lh1, float lh2, float lh3, void *t_v0, void *t_v1, void *t_v2, void *t_v3, void *t_v4)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v3)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh2 * ((float *) t_v3)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh3 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v4)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v4)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh2 * ((float *) t_v3)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v4)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh1 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh2 * ((float *) t_v3)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v4)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh7s (float lh1, float lh2, float lh3, void *t_v0, void *t_v1, void *t_v2, void *t_v3)
{

  int iflag, i;
  float temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh3 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v3)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v3)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v3)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i] + lh3 * ((float *) t_v3)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taccumh8s (float lh1, float lh2, void *t_v0, void *t_v1, void *t_v2)
{

  int iflag, i;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + lh2 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] * ((float *) t_v2)[(0) * hess_maxsize_1 + i];
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_1; i++)
	    {
	      ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_v0)[(0 + 1) * hess_maxsize_1 + i] + (0.5 * lh1 * ((float *) t_v1)[(0) * hess_maxsize_1 + i] + lh2 * ((float *) t_v2)[(0) * hess_maxsize_1 + i]) * ((float *) t_v1)[(0) * hess_maxsize_1 + i];
	    }
	}
      }
  }
}


void 
ad_fh_taddad (double c, void *t_c, double a, void *t_a, double b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = ((double *) t_a)[(0) * hess_maxsize_1 + i] + ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taddbd (double c, void *t_c, double a, void *t_a, double b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = ((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taddcd (double c, void *t_c, double a, double b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tsubad (double c, void *t_c, double a, void *t_a, double b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = ((double *) t_a)[(0) * hess_maxsize_1 + i] - ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] - ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tsubbd (double c, void *t_c, double a, void *t_a, double b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = ((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tsubcd (double c, void *t_c, double a, double b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = -((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = -((double *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmulad (double c, void *t_c, double a, void *t_a, double b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = b * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + a * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i] + ((double *) t_a)[(0) * hess_maxsize_1 + i] * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = b * ((double *) t_a)[(0) * hess_maxsize_1 + i] + a * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmulbd (double c, void *t_c, double a, void *t_a, double b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = b * ((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = b * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmulcd (double c, void *t_c, double a, double b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = a * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = a * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tdivad (double c, void *t_c, double a, void *t_a, double b, void *t_b)
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = temp2 * (((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] - temp1 * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i] - temp2 * ((double *) t_a)[(0) * hess_maxsize_1 + i] * ((double *) t_b)[(0) * hess_maxsize_1 + i] + temp3 * ((double *) t_b)[(0) * hess_maxsize_1 + i] * ((double *) t_b)[(0) * hess_maxsize_1 + i]);
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = temp2 * (((double *) t_a)[(0) * hess_maxsize_1 + i] - temp1 * ((double *) t_b)[(0) * hess_maxsize_1 + i]);
      }
  }
}


void 
ad_fh_tdivbd (double c, void *t_c, double a, void *t_a, double b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = ((double *) t_a)[(0) * hess_maxsize_1 + i] / b;
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] / b;
      }
  }
}


void 
ad_fh_tdivcd (double c, void *t_c, double a, double b, void *t_b)
{

  int i;
  double temp1, temp2;

  {
    temp1 = -a / (b * b);
    temp2 = -temp1 / b;
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = temp1 * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i] + temp2 * ((double *) t_b)[(0) * hess_maxsize_1 + i] * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = temp1 * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tcopyd (double c, void *t_c, double a, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = ((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmind (double c, void *t_c, double a, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = -((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = -((double *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tzerod (double c, void *t_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = 0.0;
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = 0.0;
      }
  }
}


void 
ad_fh_taddas (float c, void *t_c, float a, void *t_a, float b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = ((float *) t_a)[(0) * hess_maxsize_1 + i] + ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tintrin1d (double fa, double faa, double c, void *t_c, double a, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = fa * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * faa * ((double *) t_a)[(0) * hess_maxsize_1 + i] * ((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = fa * ((double *) t_a)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tintrin2d (double fa, double fb, double faa, double fab, double fbb, double c, void *t_c, double a, void *t_a, double b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0 + 1) * hess_maxsize_1 + i] = fa * ((double *) t_a)[(0 + 1) * hess_maxsize_1 + i] + fb * ((double *) t_b)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * faa * ((double *) t_a)[(0) * hess_maxsize_1 + i] * ((double *) t_a)[(0) * hess_maxsize_1 + i] + 0.5 * fbb * ((double *) t_b)[(0) * hess_maxsize_1 + i] * ((double *) t_b)[(0) * hess_maxsize_1 + i] + fab * ((double *) t_a)[(0) * hess_maxsize_1 + i] * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((double *) t_c)[(0) * hess_maxsize_1 + i] = fa * ((double *) t_a)[(0) * hess_maxsize_1 + i] + fb * ((double *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taddbs (float c, void *t_c, float a, void *t_a, float b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = ((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_taddcs (float c, void *t_c, float a, float b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tsubas (float c, void *t_c, float a, void *t_a, float b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = ((float *) t_a)[(0) * hess_maxsize_1 + i] - ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] - ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tsubbs (float c, void *t_c, float a, void *t_a, float b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = ((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tsubcs (float c, void *t_c, float a, float b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = -((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = -((float *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmulas (float c, void *t_c, float a, void *t_a, float b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = b * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + a * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i] + ((float *) t_a)[(0) * hess_maxsize_1 + i] * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = b * ((float *) t_a)[(0) * hess_maxsize_1 + i] + a * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmulbs (float c, void *t_c, float a, void *t_a, float b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = b * ((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = b * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmulcs (float c, void *t_c, float a, float b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = a * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = a * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tdivas (float c, void *t_c, float a, void *t_a, float b, void *t_b)
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = temp2 * (((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] - temp1 * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i] - temp2 * ((float *) t_a)[(0) * hess_maxsize_1 + i] * ((float *) t_b)[(0) * hess_maxsize_1 + i] + temp3 * ((float *) t_b)[(0) * hess_maxsize_1 + i] * ((float *) t_b)[(0) * hess_maxsize_1 + i]);
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = temp2 * (((float *) t_a)[(0) * hess_maxsize_1 + i] - temp1 * ((float *) t_b)[(0) * hess_maxsize_1 + i]);
      }
  }
}


void 
ad_fh_tdivbs (float c, void *t_c, float a, void *t_a, float b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = ((float *) t_a)[(0) * hess_maxsize_1 + i] / b;
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] / b;
      }
  }
}


void 
ad_fh_tdivcs (float c, void *t_c, float a, float b, void *t_b)
{

  int i;
  float temp1, temp2;

  {
    temp1 = -a / (b * b);
    temp2 = -temp1 / b;
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = temp1 * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i] + temp2 * ((float *) t_b)[(0) * hess_maxsize_1 + i] * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = temp1 * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tcopys (float c, void *t_c, float a, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = ((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tmins (float c, void *t_c, float a, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = -((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = -((float *) t_a)[(0 + 1) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tzeros (float c, void *t_c)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = 0.0;
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = 0.0;
      }
  }
}


void 
ad_fh_tintrin1s (float fa, float faa, float c, void *t_c, float a, void *t_a)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = fa * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * faa * ((float *) t_a)[(0) * hess_maxsize_1 + i] * ((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = fa * ((float *) t_a)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_fh_tintrin2s (float fa, float fb, float faa, float fab, float fbb, float c, void *t_c, float a, void *t_a, float b, void *t_b)
{

  int i;

  {
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0 + 1) * hess_maxsize_1 + i] = fa * ((float *) t_a)[(0 + 1) * hess_maxsize_1 + i] + fb * ((float *) t_b)[(0 + 1) * hess_maxsize_1 + i] + 0.5 * faa * ((float *) t_a)[(0) * hess_maxsize_1 + i] * ((float *) t_a)[(0) * hess_maxsize_1 + i] + 0.5 * fbb * ((float *) t_b)[(0) * hess_maxsize_1 + i] * ((float *) t_b)[(0) * hess_maxsize_1 + i] + fab * ((float *) t_a)[(0) * hess_maxsize_1 + i] * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
    for (i = 0; i < hess_size_1; i++)
      {
	((float *) t_c)[(0) * hess_maxsize_1 + i] = fa * ((float *) t_a)[(0) * hess_maxsize_1 + i] + fb * ((float *) t_b)[(0) * hess_maxsize_1 + i];
      }
  }
}


void 
ad_h_fpadda0d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j] + lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpadda1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpadda2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpadda3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
  }
}


void 
ad_h_fpaddb1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpaddb3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
  }
}


void 
ad_h_fpaddc2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpaddc3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_b[i];
      }
  }
}


void 
ad_h_fpsuba0d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j] - lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpsuba1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpsuba2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpsuba3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
  }
}


void 
ad_h_fpsubb1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpsubb3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
  }
}


void 
ad_h_fpsubc2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpsubc3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_b[i];
      }
  }
}


void 
ad_h_fpmula0d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = b * lh_a[i][j] + a * lh_b[i][j] + lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmula1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = b * lh_a[i][j] + lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmula2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = a * lh_b[i][j] + lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmula3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmulb1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = b * lh_a[i][j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i];
      }
  }
}


void 
ad_h_fpmulb3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i];
      }
  }
}


void 
ad_h_fpmulc2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = a * lh_b[i][j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = a * lg_b[i];
      }
  }
}


void 
ad_h_fpmulc3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = a * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva0d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (lh_a[i][j] - temp1 * lh_b[i][j] + temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (lh_a[i][j] + temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (-temp1 * lh_b[i][j] + temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdivb1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j] / b;
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] / b;
      }
  }
}


void 
ad_h_fpdivb3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] / b;
      }
  }
}


void 
ad_h_fpdivc2d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;
  double temp1, temp2, temp3;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp1 * lh_b[i][j] - temp2 * lg_b[i] * lg_b[j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp1 * lg_b[i];
      }
  }
}


void 
ad_h_fpdivc3d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double b, double lg_b[5], double lh_b[5][5])
{

  int j, i;
  double temp1, temp2;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -temp2 * lg_b[i] * lg_b[j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp1 * lg_b[i];
      }
  }
}


void 
ad_h_fpinitd (int ldg, int len, int n, double lg_c[5], double lh_c[5][5])
{

  int i;

  {
    n = n - 1;
    for (i = 0; i < len; i++)
      {
	lg_c[i] = 0.0;
      }
    lg_c[n] = 1.0;
  }
}


void 
ad_h_fpmin0d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_a[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpmin1d (int ldg, int len, double c, double lg_c[5], double lh_c[5][5], double a, double lg_a[5], double lh_a[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_a[i];
      }
  }
}


void 
ad_h_fpzerod (int ldg, int len, double lg_c[5], double lh_c[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = 0.0;
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = 0.0;
	  }
      }
  }
}


void 
ad_h_fpadda0s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j] + lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpadda1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpadda2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpadda3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] + lg_b[i];
      }
  }
}


void 
ad_h_fpaddb1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpaddb3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
  }
}


void 
ad_h_fpaddc2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpaddc3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_b[i];
      }
  }
}


void 
ad_h_fpsuba0s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j] - lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpsuba1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpsuba2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpsuba3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] - lg_b[i];
      }
  }
}


void 
ad_h_fpsubb1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpsubb3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i];
      }
  }
}


void 
ad_h_fpsubc2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_b[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -lh_b[i][j];
	  }
      }
  }
}


void 
ad_h_fpsubc3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_b[i];
      }
  }
}


void 
ad_h_fpmula0s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = b * lh_a[i][j] + a * lh_b[i][j] + lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmula1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = b * lh_a[i][j] + lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmula2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = a * lh_b[i][j] + lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmula3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i] + a * lg_b[i];
      }
  }
}


void 
ad_h_fpmulb1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = b * lh_a[i][j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i];
      }
  }
}


void 
ad_h_fpmulb3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = b * lg_a[i];
      }
  }
}


void 
ad_h_fpmulc2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = a * lh_b[i][j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = a * lg_b[i];
      }
  }
}


void 
ad_h_fpmulc3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = a * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva0s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (lh_a[i][j] - temp1 * lh_b[i][j] + temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (lh_a[i][j] + temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (-temp1 * lh_b[i][j] + temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdiva3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp3 * (temp2 * lg_b[i] * lg_b[j] - temp3 * (lg_a[i] * lg_b[j] + lg_a[j] * lg_b[i]));
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp3 * lg_a[i] - temp1 * temp3 * lg_b[i];
      }
  }
}


void 
ad_h_fpdivb1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = lh_a[i][j] / b;
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] / b;
      }
  }
}


void 
ad_h_fpdivb3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = lg_a[i] / b;
      }
  }
}


void 
ad_h_fpdivc2s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;
  float temp1, temp2, temp3;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = temp1 * lh_b[i][j] - temp2 * lg_b[i] * lg_b[j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp1 * lg_b[i];
      }
  }
}


void 
ad_h_fpdivc3s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float b, float lg_b[5], float lh_b[5][5])
{

  int j, i;
  float temp1, temp2;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -temp2 * lg_b[i] * lg_b[j];
	  }
      }
    for (i = 0; i < len; i++)
      {
	lg_c[i] = temp1 * lg_b[i];
      }
  }
}


void 
ad_h_fpinits (int ldg, int len, int n, float lg_c[5], float lh_c[5][5])
{

  int i;

  {
    n = n - 1;
    for (i = 0; i < len; i++)
      {
	lg_c[i] = 0.0;
      }
    lg_c[n] = 1.0;
  }
}


void 
ad_h_fpmin0s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_a[i];
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = -lh_a[i][j];
	  }
      }
  }
}


void 
ad_h_fpmin1s (int ldg, int len, float c, float lg_c[5], float lh_c[5][5], float a, float lg_a[5], float lh_a[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = -lg_a[i];
      }
  }
}


void 
ad_h_fpzeros (int ldg, int len, float lg_c[5], float lh_c[5][5])
{

  int j, i;

  {
    for (i = 0; i < len; i++)
      {
	lg_c[i] = 0.0;
      }
    for (i = 0; i < len; i++)
      {
	for (j = 0; j < i + (1 - 0); j++)
	  {
	    lh_c[i][j] = 0.0;
	  }
      }
  }
}


void 
ad_h_tpadda0d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i] + lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpadda1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpadda2d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpadda3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
  }
}


void 
ad_h_tpaddb1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpaddb3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
  }
}


void 
ad_h_tpaddc2d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpaddc3d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_b[0][i];
      }
  }
}


void 
ad_h_tpsuba0d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i] - lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpsuba1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpsuba2d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpsuba3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
  }
}


void 
ad_h_tpsubb1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpsubb3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
  }
}


void 
ad_h_tpsubc2d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpsubc3d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula0d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = b * lt_a[0 + 1][i] + a * lt_b[0 + 1][i] + lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = b * lt_a[0 + 1][i] + lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula2d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = a * lt_b[0 + 1][i] + lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmulb1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = b * lt_a[0 + 1][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i];
      }
  }
}


void 
ad_h_tpmulb3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i];
      }
  }
}


void 
ad_h_tpmulc2d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = a * lt_b[0 + 1][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmulc3d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpdiva0d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (lt_a[0 + 1][i] - temp1 * lt_b[0 + 1][i] - temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdiva1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (lt_a[0 + 1][i] - temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdiva2d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (-temp1 * lt_b[0 + 1][i] - temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdiva3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b, double lt_b[2][15])
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (-temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdivb1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * lt_a[0 + 1][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * lt_a[0][i];
      }
  }
}


void 
ad_h_tpdivb3d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15], double b)
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * lt_a[0][i];
      }
  }
}


void 
ad_h_tpdivc2d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;
  double temp1, temp2, temp3;

  {
    temp1 = a / (b * b);
    temp2 = temp1 / b;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -temp1 * lt_b[0 + 1][i] + temp2 * lt_b[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -temp1 * lt_b[0][i];
      }
  }
}


void 
ad_h_tpdivc3d (int ldg, int len, double c, double lt_c[2][15], double a, double b, double lt_b[2][15])
{

  int i;
  double temp1, temp2;

  {
    temp1 = a / (b * b);
    temp2 = temp1 / b;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * lt_b[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -temp1 * lt_b[0][i];
      }
  }
}


void 
ad_h_tpinit1d (int ldg, int len, double lt_c[2][15], int n1)
{

  int i;

  {
    n1 = n1 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
  }
}


void 
ad_h_tpinit2d (int ldg, int len, double lt_c[2][15], int n1, int n2)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
  }
}


void 
ad_h_tpinit3d (int ldg, int len, double lt_c[2][15], int n1, int n2, int n3)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    n3 = n3 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
    lt_c[0][n3] = 1.0;
  }
}


void 
ad_h_tpinit4d (int ldg, int len, double lt_c[2][15], int n1, int n2, int n3, int n4)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    n3 = n3 - 1;
    n4 = n4 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
    lt_c[0][n3] = 1.0;
    lt_c[0][n4] = 1.0;
  }
}


void 
ad_h_tpinit5d (int ldg, int len, double lt_c[2][15], int n1, int n2, int n3, int n4, int n5)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    n3 = n3 - 1;
    n4 = n4 - 1;
    n5 = n5 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
    lt_c[0][n3] = 1.0;
    lt_c[0][n4] = 1.0;
    lt_c[0][n5] = 1.0;
  }
}


void 
ad_h_tpmin0d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_a[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpmin1d (int ldg, int len, double c, double lt_c[2][15], double a, double lt_a[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_a[0][i];
      }
  }
}


void 
ad_h_tpzerod (int ldg, int len, double lt_c[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = 0.0;
      }
  }
}


void 
ad_h_tpadda0s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i] + lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpadda1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpadda2s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpadda3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] + lt_b[0][i];
      }
  }
}


void 
ad_h_tpaddb1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpaddb3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
  }
}


void 
ad_h_tpaddc2s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpaddc3s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_b[0][i];
      }
  }
}


void 
ad_h_tpsuba0s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i] - lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpsuba1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpsuba2s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpsuba3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i] - lt_b[0][i];
      }
  }
}


void 
ad_h_tpsubb1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpsubb3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = lt_a[0][i];
      }
  }
}


void 
ad_h_tpsubc2s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -lt_b[0 + 1][i];
      }
  }
}


void 
ad_h_tpsubc3s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula0s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = b * lt_a[0 + 1][i] + a * lt_b[0 + 1][i] + lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = b * lt_a[0 + 1][i] + lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula2s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = a * lt_b[0 + 1][i] + lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmula3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = lt_a[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i] + a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmulb1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = b * lt_a[0 + 1][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i];
      }
  }
}


void 
ad_h_tpmulb3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = b * lt_a[0][i];
      }
  }
}


void 
ad_h_tpmulc2s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = a * lt_b[0 + 1][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpmulc3s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = a * lt_b[0][i];
      }
  }
}


void 
ad_h_tpdiva0s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (lt_a[0 + 1][i] - temp1 * lt_b[0 + 1][i] - temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdiva1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (lt_a[0 + 1][i] - temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdiva2s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (-temp1 * lt_b[0 + 1][i] - temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdiva3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b, float lt_b[2][15])
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * (-temp2 * lt_a[0][i] * lt_b[0][i] + temp3 * lt_b[0][i] * lt_b[0][i]);
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * (lt_a[0][i] - temp1 * lt_b[0][i]);
      }
  }
}


void 
ad_h_tpdivb1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * lt_a[0 + 1][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * lt_a[0][i];
      }
  }
}


void 
ad_h_tpdivb3s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15], float b)
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 1.0 / b;
    temp3 = temp1 * temp2;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = temp2 * lt_a[0][i];
      }
  }
}


void 
ad_h_tpdivc2s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;
  float temp1, temp2, temp3;

  {
    temp1 = a / (b * b);
    temp2 = temp1 / b;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -temp1 * lt_b[0 + 1][i] + temp2 * lt_b[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -temp1 * lt_b[0][i];
      }
  }
}


void 
ad_h_tpdivc3s (int ldg, int len, float c, float lt_c[2][15], float a, float b, float lt_b[2][15])
{

  int i;
  float temp1, temp2;

  {
    temp1 = a / (b * b);
    temp2 = temp1 / b;
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = temp2 * lt_b[0][i] * lt_b[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -temp1 * lt_b[0][i];
      }
  }
}


void 
ad_h_tpinit1s (int ldg, int len, float lt_c[2][15], int n1)
{

  int i;

  {
    n1 = n1 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
  }
}


void 
ad_h_tpinit2s (int ldg, int len, float lt_c[2][15], int n1, int n2)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
  }
}


void 
ad_h_tpinit3s (int ldg, int len, float lt_c[2][15], int n1, int n2, int n3)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    n3 = n3 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
    lt_c[0][n3] = 1.0;
  }
}


void 
ad_h_tpinit4s (int ldg, int len, float lt_c[2][15], int n1, int n2, int n3, int n4)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    n3 = n3 - 1;
    n4 = n4 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
    lt_c[0][n3] = 1.0;
    lt_c[0][n4] = 1.0;
  }
}


void 
ad_h_tpinit5s (int ldg, int len, float lt_c[2][15], int n1, int n2, int n3, int n4, int n5)
{

  int i;

  {
    n1 = n1 - 1;
    n2 = n2 - 1;
    n3 = n3 - 1;
    n4 = n4 - 1;
    n5 = n5 - 1;
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    lt_c[0][n1] = 1.0;
    lt_c[0][n2] = 1.0;
    lt_c[0][n3] = 1.0;
    lt_c[0][n4] = 1.0;
    lt_c[0][n5] = 1.0;
  }
}


void 
ad_h_tpmin0s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_a[0][i];
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = -lt_a[0 + 1][i];
      }
  }
}


void 
ad_h_tpmin1s (int ldg, int len, float c, float lt_c[2][15], float a, float lt_a[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = -lt_a[0][i];
      }
  }
}


void 
ad_h_tpzeros (int ldg, int len, float lt_c[2][15])
{

  int i;

  {
    for (i = 0; i < len; i++)
      {
	lt_c[0][i] = 0.0;
      }
    for (i = 0; i < len; i++)
      {
	lt_c[0 + 1][i] = 0.0;
      }
  }
}


void 
ad_ph_faddad (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b, double *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] + g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = ((double *) h_a)[(i) * hess_maxsize_1 + j] + ((double *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_faddbd (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = ((double *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_faddcd (double c, double *g_c, void *h_c, double a, double b, double *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = ((double *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fsubad (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b, double *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] - g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = ((double *) h_a)[(i) * hess_maxsize_1 + j] - ((double *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fsubbd (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = ((double *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fsubcd (double c, double *g_c, void *h_c, double a, double b, double *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = -g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = -((double *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fmulad (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b, double *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = b * ((double *) h_a)[(i) * hess_maxsize_1 + j] + a * ((double *) h_b)[(i) * hess_maxsize_1 + j] + g_a[hess_size_1 + i] * g_b[j] + g_a[j] * g_b[hess_size_1 + i];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = b * g_a[i] + a * g_b[i];
      }
  }
}


void 
ad_ph_fmulbd (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = b * g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = b * ((double *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fmulcd (double c, double *g_c, void *h_c, double a, double b, double *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = a * g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = a * ((double *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fdivad (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b, double *g_b, void *h_b)
{

  int j, i;
  double temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = temp3 * (((double *) h_a)[(i) * hess_maxsize_1 + j] - temp1 * ((double *) h_b)[(i) * hess_maxsize_1 + j] + temp2 * g_b[hess_size_1 + i] * g_b[j] - temp3 * (g_a[hess_size_1 + i] * g_b[j] + g_a[j] * g_b[hess_size_1 + i]));
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] / b - temp1 * temp3 * g_b[i];
      }
  }
}


void 
ad_ph_fdivbd (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] / b;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = ((double *) h_a)[(i) * hess_maxsize_1 + j] / b;
	  }
      }
  }
}


void 
ad_ph_fdivcd (double c, double *g_c, void *h_c, double a, double b, double *g_b, void *h_b)
{

  int j, i;
  double temp1, temp2;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = temp1 * ((double *) h_b)[(i) * hess_maxsize_1 + j] - temp2 * g_b[hess_size_1 + i] * g_b[j];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = temp1 * g_b[i];
      }
  }
}


void 
ad_ph_fcopyd (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = ((double *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fmind (double c, double *g_c, void *h_c, double a, double *g_a, void *h_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = -g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = -((double *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fzerod (double c, double *g_c, void *h_c)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = 0.0;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = 0.0;
	  }
      }
  }
}


void 
ad_ph_fintrin1d (double fa, double faa, double c, double *g_c, void *h_c, double a, double *g_a, void *h_a)
{

  int j, i;
  double temp;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	temp = faa * g_a[hess_size_1 + i];
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = fa * ((double *) h_a)[(i) * hess_maxsize_1 + j] + temp * g_a[j];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = fa * g_a[i];
      }
  }
}


void 
ad_ph_fintrin2d (double fa, double fb, double faa, double fab, double fbb, double c, double *g_c, void *h_c, double a, double *g_a, void *h_a, double b, double *g_b, void *h_b)
{

  int j, i;
  double temp1, temp2;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	temp1 = faa * g_a[hess_size_1 + i] + fab * g_b[hess_size_1 + i];
	temp2 = fab * g_a[hess_size_1 + i] + fbb * g_b[hess_size_1 + i];
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = fa * ((double *) h_a)[(i) * hess_maxsize_1 + j] + fb * ((double *) h_b)[(i) * hess_maxsize_1 + j] + temp1 * g_a[j] + temp2 * g_b[j];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = fa * g_a[i] + fb * g_b[i];
      }
  }
}


void 
ad_ph_faddas (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b, float *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] + g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = ((float *) h_a)[(i) * hess_maxsize_1 + j] + ((float *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_faddbs (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = ((float *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_faddcs (float c, float *g_c, void *h_c, float a, float b, float *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = ((float *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fsubas (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b, float *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] - g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = ((float *) h_a)[(i) * hess_maxsize_1 + j] - ((float *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fsubbs (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = ((float *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fsubcs (float c, float *g_c, void *h_c, float a, float b, float *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = -g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = -((float *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fmulas (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b, float *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = b * ((float *) h_a)[(i) * hess_maxsize_1 + j] + a * ((float *) h_b)[(i) * hess_maxsize_1 + j] + g_a[hess_size_1 + i] * g_b[j] + g_a[j] * g_b[hess_size_1 + i];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = b * g_a[i] + a * g_b[i];
      }
  }
}


void 
ad_ph_fmulbs (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = b * g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = b * ((float *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fmulcs (float c, float *g_c, void *h_c, float a, float b, float *g_b, void *h_b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = a * g_b[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = a * ((float *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fdivas (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b, float *g_b, void *h_b)
{

  int j, i;
  float temp1, temp2, temp3;

  {
    temp1 = a / b;
    temp2 = 2.0 * temp1 / b;
    temp3 = 1.0 / b;
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = temp3 * (((float *) h_a)[(i) * hess_maxsize_1 + j] - temp1 * ((float *) h_b)[(i) * hess_maxsize_1 + j] + temp2 * g_b[hess_size_1 + i] * g_b[j] - temp3 * (g_a[hess_size_1 + i] * g_b[j] + g_a[j] * g_b[hess_size_1 + i]));
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] / b - temp1 * temp3 * g_b[i];
      }
  }
}


void 
ad_ph_fdivbs (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i] / b;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = ((float *) h_a)[(i) * hess_maxsize_1 + j] / b;
	  }
      }
  }
}


void 
ad_ph_fdivcs (float c, float *g_c, void *h_c, float a, float b, float *g_b, void *h_b)
{

  int j, i;
  float temp1, temp2;

  {
    temp1 = -a / (b * b);
    temp2 = 2.0 * temp1 / b;
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = temp1 * ((float *) h_b)[(i) * hess_maxsize_1 + j] - temp2 * g_b[hess_size_1 + i] * g_b[j];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = temp1 * g_b[i];
      }
  }
}


void 
ad_ph_fcopys (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = ((float *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fmins (float c, float *g_c, void *h_c, float a, float *g_a, void *h_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = -g_a[i];
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = -((float *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_fzeros (float c, float *g_c, void *h_c)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = 0.0;
      }
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = 0.0;
	  }
      }
  }
}


void 
ad_ph_fintrin1s (float fa, float faa, float c, float *g_c, void *h_c, float a, float *g_a, void *h_a)
{

  int j, i;
  float temp;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	temp = faa * g_a[hess_size_1 + i];
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = fa * ((float *) h_a)[(i) * hess_maxsize_1 + j] + temp * g_a[j];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = fa * g_a[i];
      }
  }
}


void 
ad_ph_fintrin2s (float fa, float fb, float faa, float fab, float fbb, float c, float *g_c, void *h_c, float a, float *g_a, void *h_a, float b, float *g_b, void *h_b)
{

  int j, i;
  float temp1, temp2;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	temp1 = faa * g_a[hess_size_1 + i] + fab * g_b[hess_size_1 + i];
	temp2 = fab * g_a[hess_size_1 + i] + fbb * g_b[hess_size_1 + i];
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = fa * ((float *) h_a)[(i) * hess_maxsize_1 + j] + fb * ((float *) h_b)[(i) * hess_maxsize_1 + j] + temp1 * g_a[j] + temp2 * g_b[j];
	  }
      }
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = fa * g_a[i] + fb * g_b[i];
      }
  }
}


void 
ad_ph_accumg1d (double lg1, double *g_c, double *g_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = lg1 * g_a[i];
      }
  }
}


void 
ad_ph_accumg2d (double lg1, double lg2, double *g_c, double *g_a, double *g_b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = lg1 * g_a[i] + lg2 * g_b[i];
      }
  }
}


void 
ad_ph_accumg3d (double lg1, double lg2, double lg3, double *g_d, double *g_a, double *g_b, double *g_c)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_d[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i];
      }
  }
}


void 
ad_ph_accumg4d (double lg1, double lg2, double lg3, double lg4, double *g_f, double *g_a, double *g_b, double *g_c, double *g_d)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_f[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i];
      }
  }
}


void 
ad_ph_accumg5d (double lg1, double lg2, double lg3, double lg4, double lg5, double *g_f, double *g_a, double *g_b, double *g_c, double *g_d, double *g_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_f[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i] + lg5 * g_e[i];
      }
  }
}


void 
ad_ph_accumhg1d (double lg1, void *h_c, void *h_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = lg1 * ((double *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg2d (double lg1, double lg2, void *h_c, void *h_a, void *h_b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_c)[(i) * hess_maxsize_1 + j] = lg1 * ((double *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((double *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg3d (double lg1, double lg2, double lg3, void *h_d, void *h_a, void *h_b, void *h_c)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_d)[(i) * hess_maxsize_1 + j] = lg1 * ((double *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((double *) h_b)[(i) * hess_maxsize_1 + j] + lg3 * ((double *) h_c)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg4d (double lg1, double lg2, double lg3, double lg4, void *h_f, void *h_a, void *h_b, void *h_c, void *h_d)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_f)[(i) * hess_maxsize_1 + j] = lg1 * ((double *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((double *) h_b)[(i) * hess_maxsize_1 + j] + lg3 * ((double *) h_c)[(i) * hess_maxsize_1 + j] + lg4 * ((double *) h_d)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg5d (double lg1, double lg2, double lg3, double lg4, double lg5, void *h_f, void *h_a, void *h_b, void *h_c, void *h_d, void *h_e)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((double *) h_f)[(i) * hess_maxsize_1 + j] = lg1 * ((double *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((double *) h_b)[(i) * hess_maxsize_1 + j] + lg3 * ((double *) h_c)[(i) * hess_maxsize_1 + j] + lg4 * ((double *) h_d)[(i) * hess_maxsize_1 + j] + lg5 * ((double *) h_e)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumh1d (double lh1, void *h_v0, double *g_v1)
{

  int j, i;
  double temp1;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_2; i++)
	  {
	    temp1 = g_v1[hess_size_1 + i] * lh1;
	    for (j = 0; j < hess_size_1; j++)
	      {
		((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
	      }
	  }
      }
  }
}


void 
ad_ph_accumh2d (double lh1, double lh2, void *h_v0, double *g_v1, double *g_v2)
{

  int iflag, j, i;
  double temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      temp2 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh3d (double lh1, double lh2, double lh3, void *h_v0, double *g_v1, double *g_v2)
{

  int iflag, j, i;
  double temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      temp2 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh3;
	      temp2 = g_v2[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j] + temp2 * g_v1[j];
		}
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2 + g_v1[hess_size_1 + i] * lh3;
	      temp2 = g_v2[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j] + temp2 * g_v1[j];
		}
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh3;
	      temp2 = g_v2[hess_size_1 + i] * lh2 + g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh4d (double lh1, void *h_v0, double *g_v1, double *g_v2)
{

  int j, i;
  double temp1, temp2;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_2; i++)
	  {
	    temp1 = g_v2[hess_size_1 + i] * lh1;
	    temp2 = g_v1[hess_size_1 + i] * lh1;
	    for (j = 0; j < hess_size_1; j++)
	      {
		((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
	      }
	  }
      }
  }
}


void 
ad_ph_accumh5d (double lh1, double lh2, void *h_v0, double *g_v1, double *g_v2, double *g_v3)
{

  int iflag, j, i;
  double temp1, temp2, temp3;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh6d (double lh1, double lh2, double lh3, void *h_v0, double *g_v1, double *g_v2, double *g_v3, double *g_v4)
{

  int iflag, j, i;
  double temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v4[j];
		}
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v4[j];
		}
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh2 + g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j] + temp3 * g_v4[j];
		}
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh2 + g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh2;
	      temp4 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j] + temp4 * g_v4[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh7d (double lh1, double lh2, double lh3, void *h_v0, double *g_v1, double *g_v2, double *g_v3)
{

  int iflag, j, i;
  double temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2 + g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh2 + g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh8d (double lh1, double lh2, void *h_v0, double *g_v1, double *g_v2)
{

  int iflag, j, i;
  double temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((double *) h_v0)[(i) * hess_maxsize_1 + j] = ((double *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumg1s (float lg1, float *g_c, float *g_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = lg1 * g_a[i];
      }
  }
}


void 
ad_ph_accumg2s (float lg1, float lg2, float *g_c, float *g_a, float *g_b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_c[i] = lg1 * g_a[i] + lg2 * g_b[i];
      }
  }
}


void 
ad_ph_accumg3s (float lg1, float lg2, float lg3, float *g_d, float *g_a, float *g_b, float *g_c)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_d[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i];
      }
  }
}


void 
ad_ph_accumg4s (float lg1, float lg2, float lg3, float lg4, float *g_f, float *g_a, float *g_b, float *g_c, float *g_d)
{

  int j, i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_f[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i];
      }
  }
}


void 
ad_ph_accumg5s (float lg1, float lg2, float lg3, float lg4, float lg5, float *g_f, float *g_a, float *g_b, float *g_c, float *g_d, float *g_e)
{

  int i;

  {
    for (i = 0; i < hess_size_1 + hess_size_2; i++)
      {
	g_f[i] = lg1 * g_a[i] + lg2 * g_b[i] + lg3 * g_c[i] + lg4 * g_d[i] + lg5 * g_e[i];
      }
  }
}


void 
ad_ph_accumhg1s (float lg1, void *h_c, void *h_a)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = lg1 * ((float *) h_a)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg2s (float lg1, float lg2, void *h_c, void *h_a, void *h_b)
{

  int n, j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_c)[(i) * hess_maxsize_1 + j] = lg1 * ((float *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((float *) h_b)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg3s (float lg1, float lg2, float lg3, void *h_d, void *h_a, void *h_b, void *h_c)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_d)[(i) * hess_maxsize_1 + j] = lg1 * ((float *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((float *) h_b)[(i) * hess_maxsize_1 + j] + lg3 * ((float *) h_c)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg4s (float lg1, float lg2, float lg3, float lg4, void *h_f, void *h_a, void *h_b, void *h_c, void *h_d)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_f)[(i) * hess_maxsize_1 + j] = lg1 * ((float *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((float *) h_b)[(i) * hess_maxsize_1 + j] + lg3 * ((float *) h_c)[(i) * hess_maxsize_1 + j] + lg4 * ((float *) h_d)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumhg5s (float lg1, float lg2, float lg3, float lg4, float lg5, void *h_f, void *h_a, void *h_b, void *h_c, void *h_d, void *h_e)
{

  int j, i;

  {
    for (i = 0; i < hess_size_2; i++)
      {
	for (j = 0; j < hess_size_1; j++)
	  {
	    ((float *) h_f)[(i) * hess_maxsize_1 + j] = lg1 * ((float *) h_a)[(i) * hess_maxsize_1 + j] + lg2 * ((float *) h_b)[(i) * hess_maxsize_1 + j] + lg3 * ((float *) h_c)[(i) * hess_maxsize_1 + j] + lg4 * ((float *) h_d)[(i) * hess_maxsize_1 + j] + lg5 * ((float *) h_e)[(i) * hess_maxsize_1 + j];
	  }
      }
  }
}


void 
ad_ph_accumh1s (float lh1, void *h_v0, float *g_v1)
{

  int j, i;
  float temp1;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_2; i++)
	  {
	    temp1 = g_v1[hess_size_1 + i] * lh1;
	    for (j = 0; j < hess_size_1; j++)
	      {
		((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
	      }
	  }
      }
  }
}


void 
ad_ph_accumh2s (float lh1, float lh2, void *h_v0, float *g_v1, float *g_v2)
{

  int iflag, j, i;
  float temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      temp2 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh3s (float lh1, float lh2, float lh3, void *h_v0, float *g_v1, float *g_v2)
{

  int iflag, j, i;
  float temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      temp2 = g_v2[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh3;
	      temp2 = g_v2[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j] + temp2 * g_v1[j];
		}
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2 + g_v1[hess_size_1 + i] * lh3;
	      temp2 = g_v2[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v2[j] + temp2 * g_v1[j];
		}
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh3;
	      temp2 = g_v2[hess_size_1 + i] * lh2 + g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh4s (float lh1, void *h_v0, float *g_v1, float *g_v2)
{

  int j, i;
  float temp1, temp2;

  {
    if (lh1 != 0.0)
      {
	for (i = 0; i < hess_size_2; i++)
	  {
	    temp1 = g_v2[hess_size_1 + i] * lh1;
	    temp2 = g_v1[hess_size_1 + i] * lh1;
	    for (j = 0; j < hess_size_1; j++)
	      {
		((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
	      }
	  }
      }
  }
}


void 
ad_ph_accumh5s (float lh1, float lh2, void *h_v0, float *g_v1, float *g_v2, float *g_v3)
{

  int iflag, j, i;
  float temp1, temp2, temp3;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh6s (float lh1, float lh2, float lh3, void *h_v0, float *g_v1, float *g_v2, float *g_v3, float *g_v4)
{

  int iflag, j, i;
  float temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v4[j];
		}
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v4[j];
		}
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh2 + g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j] + temp3 * g_v4[j];
		}
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh2 + g_v4[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh1;
	      temp3 = g_v1[hess_size_1 + i] * lh2;
	      temp4 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j] + temp4 * g_v4[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh7s (float lh1, float lh2, float lh3, void *h_v0, float *g_v1, float *g_v2, float *g_v3)
{

  int iflag, j, i;
  float temp1, temp2, temp3, temp4;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    if (lh3 != 0.0)
      {
	iflag = iflag + 4;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 4:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 5:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v3[j];
		}
	    }
	}
      case 6:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2 + g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      case 7:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh2 + g_v3[hess_size_1 + i] * lh3;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      temp3 = g_v1[hess_size_1 + i] * lh3;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j] + temp3 * g_v3[j];
		}
	    }
	}
      }
  }
}


void 
ad_ph_accumh8s (float lh1, float lh2, void *h_v0, float *g_v1, float *g_v2)
{

  int iflag, j, i;
  float temp1, temp2;

  {
    iflag = 0;
    if (lh1 != 0.0)
      {
	iflag = 1;
      }
    if (lh2 != 0.0)
      {
	iflag = iflag + 2;
      }
    switch (iflag)
      {
      case 1:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j];
		}
	    }
	}
      case 2:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      case 3:
	{
	  for (i = 0; i < hess_size_2; i++)
	    {
	      temp1 = g_v1[hess_size_1 + i] * lh1 + g_v2[hess_size_1 + i] * lh2;
	      temp2 = g_v1[hess_size_1 + i] * lh2;
	      for (j = 0; j < hess_size_1; j++)
		{
		  ((float *) h_v0)[(i) * hess_maxsize_1 + j] = ((float *) h_v0)[(i) * hess_maxsize_1 + j] + temp1 * g_v1[j] + temp2 * g_v2[j];
		}
	    }
	}
      }
  }
}
